from collections import deque, namedtuple
from heapq import heappop, heappush

import numpy as np
import time

from .utils import INF, elapsed_time

# https://github.mit.edu/caelan/lis-openrave/tree/master/manipulation/motion
# https://github.mit.edu/caelan/lis-openrave/commit/4d8683407ec79a7c39dab62d6779804730ff598d

Node = namedtuple('Node', ['g', 'parent'])


def retrace(visited, q):
    if q is None:
        return []
    return retrace(visited, visited[tuple(q)].parent) + [q]


def bfs(start, goal, neighbors_fn, collision_fn, max_iterations=INF, max_time=INF):
    """
    :param start: Start configuration - conf
    :param goal: End configuration - conf
    :param collision_fn: Collision function - collision_fn(q)->bool
    :param max_time: Maximum runtime - float
    :return: Path [q', ..., q"] or None if unable to find a solution
    """
    start_time = time.time()
    if collision_fn(start) or collision_fn(goal):
        return None
    iterations = 0
    visited = {tuple(start): Node(g=0, parent=None)}
    queue = deque([start])
    while queue and (iterations < max_iterations) and (elapsed_time(start_time) < max_time):
        iterations += 1
        current = queue.popleft()
        if goal is not None and tuple(current) == tuple(goal):
            return retrace(visited, current)
        for new in neighbors_fn(current):
            # TODO - make edges for real (and store bad edges)
            if (tuple(new) not in visited) and not collision_fn(new):
                visited[tuple(new)] = Node(visited[tuple(current)].g + 1, current)
                queue.append(new)
    return None

##################################################

def weighted(weight=1.):
    if weight == INF:
        return lambda g, h: h
    return lambda g, h: g + weight*h

uniform = weighted(0)
astar = weighted(1)
wastar2 = weighted(2)
wastar3 = weighted(2)
greedy = weighted(INF)
lexicographic = lambda g, h: (h, g)

def best_first(start, goal, distance_fn, neighbors_fn, collision_fn,
               max_iterations=INF, max_time=INF, priority=lexicographic):  # TODO - put start and goal in neighbors_fn
    """
    :param start: Start configuration - conf
    :param goal: End configuration - conf
    :param distance_fn: Distance function - distance_fn(q1, q2)->float
    :param collision_fn: Collision function - collision_fn(q)->bool
    :param max_time: Maximum runtime - float
    :return: Path [q', ..., q"] or None if unable to find a solution
    """
    start_time = time.time()
    if collision_fn(start) or collision_fn(goal):
        return None
    queue = [(priority(0, distance_fn(start, goal)), 0, start)]
    visited = {tuple(start): Node(g=0, parent=None)}
    iterations = 0
    while queue and (iterations < max_iterations) and (elapsed_time(start_time) < max_time):
        _, current_g, current = heappop(queue)
        current = np.array(current)
        if visited[tuple(current)].g != current_g:
            continue
        # TODO: lazy collision_fn checking
        iterations += 1
        if tuple(current) == tuple(goal):
            return retrace(visited, current)
        for new in neighbors_fn(current):
            new_g = current_g + distance_fn(current, new)
            if (tuple(new) not in visited or new_g < visited[tuple(new)].g) and not collision_fn(new):
                visited[tuple(new)] = Node(new_g, current)
                # ValueError: The truth value of an array with more than one
                # element is ambiguous.
                heappush(queue, (priority(new_g, distance_fn(new, goal)), new_g, new))
    return None



"""
file I/O
"""

import os
import pickle
import json
import datetime
from .shared_const import DATE_FORMAT

SEPARATOR = '\n' + 50*'-' + '\n'

#def inf_generator():
#    return iter(int, 1)
# inf_generator = count

def print_separator(n=50):
    print('\n' + n*'-' + '\n')

def read(filename):
    with open(filename, 'r') as f:
        return f.read()

def write(filename, string):
    with open(filename, 'w') as f:
        f.write(string)

def read_pickle(filename):
    """[summary]

    Parameters
    ----------
    filename : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    # Can sometimes read pickle3 from python2 by calling twice
    # Can possibly read pickle2 from python3 by using encoding='latin1'
    with open(filename, 'rb') as f:
        return pickle.load(f)

def write_pickle(filename, data):  # NOTE - cannot pickle lambda or nested functions
    """[summary]
    """
    with open(filename, 'wb') as f:
        pickle.dump(data, f)

def read_json(path):
    return json.loads(read(path))

def write_json(path, data):
    with open(path, 'w') as f:
        json.dump(data, f, indent=2, sort_keys=True)

def safe_remove(p):
    if os.path.exists(p):
        os.remove(p)

def ensure_dir(f):
    d = os.path.dirname(f)
    if not os.path.exists(d):
        os.makedirs(d)

def get_date():
    return datetime.datetime.now().strftime(DATE_FORMAT)

def load_yaml(path):
    import yaml
    # grep -r --include="*.py" "yaml\." *
    # yaml.dump()
    with open(path, 'r') as f:
        try:
            return yaml.safe_load(f)
        except yaml.YAMLError as exc:
            raise exc



import numpy as np
from random import shuffle
from itertools import islice
from collections import deque
import time
from pybullet_planning.interfaces.env_manager.pose_transformation import get_difference, get_unit_vector

__all__ = [
    'compute_path_cost',
]

INF = float('inf')

# default RRT configurations
RRT_ITERATIONS = 20
RRT_RESTARTS = 2
RRT_SMOOTHING = 20


def irange(start, stop=None, step=1):  # np.arange
    if stop is None:
        stop = start
        start = 0
    while start < stop:
        yield start
        start += step


def negate(test):
    return lambda *args, **kwargs: not test(*args, **kwargs)


def argmin(function, sequence):
    # TODO: use min
    values = list(sequence)
    scores = [function(x) for x in values]
    return values[scores.index(min(scores))]


def get_pairs(lst):
    return zip(lst[:-1], lst[1:])


def merge_dicts(*args):
    result = {}
    for d in args:
        result.update(d)
    return result
    # return dict(reduce(operator.add, [d.items() for d in args]))


def flatten(iterable_of_iterables):
    return (item for iterables in iterable_of_iterables for item in iterables)


def randomize(sequence):
    shuffle(sequence)
    return sequence


def bisect(sequence):
    sequence = list(sequence)
    indices = set()
    queue = deque([(0, len(sequence)-1)])
    while queue:
        lower, higher = queue.popleft()
        if lower > higher:
            continue
        index = int((lower + higher) / 2.)
        assert index not in indices
        #if is_even(higher - lower):
        yield sequence[index]
        queue.extend([
            (lower, index-1),
            (index+1, higher),
        ])


def take(iterable, n=INF):
    if n == INF:
        n = None  # NOTE - islice takes None instead of INF
    elif n == None:
        n = 0  # NOTE - for some of the uses
    return islice(iterable, n)


def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    enums['names'] = sorted(enums.keys(), key=lambda k: enums[k])
    return type('Enum', (), enums)

def elapsed_time(start_time):
    return time.time() - start_time

######################################

def inf_sequence():
    return iter(int, 1)


def compute_path_cost(path, cost_fn):
    """compute the total cost of a given path using the given cost function

    Parameters
    ----------
    path :
        input path
    cost_fn : function handle
        cost function - ``cost_fn(q)->float``

    Returns
    -------
    float
        total cost
    """
    if path is None:
        return INF
    return sum(cost_fn(*pair) for pair in get_pairs(path))


def remove_redundant(path, tolerance=1e-3):
    """Remove redundant waypoints in a given path if consecutive points are too close under L2 norm.
    """
    assert path
    new_path = [path[0]]
    for conf in path[1:]:
        difference = np.array(new_path[-1]) - np.array(conf)
        if not np.allclose(np.zeros(len(difference)), difference, atol=tolerance, rtol=0):
            new_path.append(conf)
    return new_path


def waypoints_from_path(path, difference_fn=None, tolerance=1e-3):
    """Remove redundant waypoints in a given path if consecutive points are too close under L2 norm.
    """
    difference_fn = difference_fn or get_difference
    path = remove_redundant(path, tolerance=tolerance)
    if len(path) < 2:
        return path
    waypoints = [path[0]]
    last_conf = path[1]
    last_difference = get_unit_vector(difference_fn(last_conf, waypoints[-1]))
    for conf in path[2:]:
        difference = get_unit_vector(difference_fn(conf, waypoints[-1]))
        if not np.allclose(last_difference, difference, atol=tolerance, rtol=0):
            waypoints.append(last_conf)
            difference = get_unit_vector(difference_fn(conf, waypoints[-1]))
        last_conf = conf
        last_difference = difference
    waypoints.append(last_conf)
    return waypoints


def convex_combination(x, y, w=0.5):
    return (1-w)*np.array(x) + w*np.array(y)

##################################################

def forward_selector(path):
    return path


def backward_selector(path):
    return reversed(list(path))


def random_selector(path):
    return randomize(path)


def bisect_selector(path):
    return bisect(path)


default_selector = bisect_selector



import os

#####################################
# TODO: clean up these data paths

DRAKE_PATH = 'models/drake/'

# Models

# Robots
MODEL_DIRECTORY = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, 'models/'))
ROOMBA_URDF = 'models/turtlebot/roomba.urdf'
TURTLEBOT_URDF = 'models/turtlebot/turtlebot_holonomic.urdf'
DRAKE_IIWA_URDF = 'models/drake/iiwa_description/urdf/iiwa14_polytope_collision.urdf'
WSG_50_URDF = 'models/drake/wsg_50_description/urdf/wsg_50_mesh_visual.urdf' # wsg_50 | wsg_50_mesh_visual | wsg_50_mesh_collision
#SCHUNK_URDF = 'models/drake/wsg_50_description/sdf/schunk_wsg_50.sdf'
PANDA_HAND_URDF = "models/franka_description/robots/hand.urdf"
PANDA_ARM_URDF = "models/franka_description/robots/panda_arm_hand.urdf"

# Pybullet Robots
#PYBULLET_DIRECTORY = add_data_path()
KUKA_IIWA_URDF = "kuka_iiwa/model.urdf"
KUKA_IIWA_GRIPPER_SDF = "kuka_iiwa/kuka_with_gripper.sdf"
R2D2_URDF = "r2d2.urdf"
MINITAUR_URDF = "quadruped/minitaur.urdf"
HUMANOID_MJCF = "mjcf/humanoid.xml"
HUSKY_URDF = "husky/husky.urdf"
RACECAR_URDF = 'racecar/racecar.urdf' # racecar_differential.urdf
PR2_GRIPPER = 'pr2_gripper.urdf'
WSG_GRIPPER = 'gripper/wsg50_one_motor_gripper.sdf' # wsg50_one_motor_gripper | wsg50_one_motor_gripper_new

# Pybullet Objects
KIVA_SHELF_SDF = "kiva_shelf/model.sdf"
FLOOR_URDF = 'plane.urdf'
TABLE_URDF = 'table'

# Objects
SMALL_BLOCK_URDF = "models/drake/objects/block_for_pick_and_place.urdf"
BLOCK_URDF = "models/drake/objects/block_for_pick_and_place_mid_size.urdf"
SINK_URDF = 'models/sink.urdf'
STOVE_URDF = 'models/stove.urdf'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import sphinx_rtd_theme

# -- General configuration ------------------------------------------------

project = 'pybullet_planning'
year = '2020'
author = 'Caelan Garrett & Yijiang Huang'
copyright = '{0}, {1}'.format(year, author)
version = release = '0.6.1'

master_doc = 'index'
source_suffix = '.rst'
# templates_path = ['_templates', ]
exclude_patterns = ['_build', '**.ipynb_checkpoints', '_notebooks']

pygments_style = 'sphinx'
show_authors = True
add_module_names = True

extlinks = {
    'issue': ('https://github.com/yijiangh/pybullet_planning/issues/%s', '#'),
    'pr': ('https://github.com/yijiangh/pybullet_planning/pull/%s', 'PR #'),
}

# -- Extension configuration ------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinx_rtd_theme',
]
if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_US'

# intersphinx options
# intersphinx_mapping = {'python': ('https://docs.python.org/', None),
#                        'roslibpy': ('http://roslibpy.readthedocs.org/en/latest/', None)}

# autodoc options
autodoc_default_options = {
    'member-order': 'bysource',
    'special-members': '__init__',
    'exclude-members': '__weakref__',
    'undoc-members': True,
    'private-members': True,
    'show-inheritance': True,
}

autodoc_member_order = 'alphabetical'

# autosummary options
autosummary_generate = True

# on_rtd is whether we are on readthedocs.org
# on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
html_theme = "sphinx_rtd_theme"
# options: https://sphinx-rtd-theme.readthedocs.io/en/latest/configuring.html

# napoleon options
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = True
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = False
napoleon_use_rtype = False

# -- Options for HTML output ----------------------------------------------

html_split_index = False
html_short_title = '%s-%s' % (project, version)
html_context = {}
html_static_path = ['_static']
html_last_updated_fmt = '%b %d, %Y'
html_copy_source = False
html_show_sourcelink = False
html_add_permalinks = ''
html_experimental_html5_writer = True
html_compact_lists = True



from collections import namedtuple
import pybullet as p

from pybullet_planning.utils import CLIENT, CIRCULAR_LIMITS, UNBOUNDED_LIMITS, INF

#####################################
# Joints

JOINT_TYPES = {
    p.JOINT_REVOLUTE: 'revolute', # 0
    p.JOINT_PRISMATIC: 'prismatic', # 1
    p.JOINT_SPHERICAL: 'spherical', # 2
    p.JOINT_PLANAR: 'planar', # 3
    p.JOINT_FIXED: 'fixed', # 4
    p.JOINT_POINT2POINT: 'point2point', # 5
    p.JOINT_GEAR: 'gear', # 6
}

def get_num_joints(body):
    return p.getNumJoints(body, physicsClientId=CLIENT)

def get_joints(body):
    return list(range(get_num_joints(body)))

def get_joint(body, joint_or_name):
    if type(joint_or_name) is str:
        return joint_from_name(body, joint_or_name)
    return joint_or_name

JointInfo = namedtuple('JointInfo', ['jointIndex', 'jointName', 'jointType',
                                     'qIndex', 'uIndex', 'flags',
                                     'jointDamping', 'jointFriction', 'jointLowerLimit', 'jointUpperLimit',
                                     'jointMaxForce', 'jointMaxVelocity', 'linkName', 'jointAxis',
                                     'parentFramePos', 'parentFrameOrn', 'parentIndex'])

def get_joint_info(body, joint):
    return JointInfo(*p.getJointInfo(body, joint, physicsClientId=CLIENT))

def get_joint_name(body, joint):
    return get_joint_info(body, joint).jointName # .decode('UTF-8')

def get_joint_names(body, joints):
    return [get_joint_name(body, joint) for joint in joints]

def joint_from_name(body, name):
    for joint in get_joints(body):
        jt_name = get_joint_name(body, joint)
        if jt_name == name or jt_name.decode('UTF-8') == name:
            return joint
    raise ValueError(body, name)

def has_joint(body, name):
    try:
        joint_from_name(body, name)
    except ValueError:
        return False
    return True

def joints_from_names(body, names):
    return tuple(joint_from_name(body, name) for name in names)

JointState = namedtuple('JointState', ['jointPosition', 'jointVelocity',
                                       'jointReactionForces', 'appliedJointMotorTorque'])

def get_joint_state(body, joint):
    return JointState(*p.getJointState(body, joint, physicsClientId=CLIENT))

def get_joint_position(body, joint):
    return get_joint_state(body, joint).jointPosition

def get_joint_velocity(body, joint):
    return get_joint_state(body, joint).jointVelocity

def get_joint_reaction_force(body, joint):
    return get_joint_state(body, joint).jointReactionForces

def get_joint_torque(body, joint):
    return get_joint_state(body, joint).appliedJointMotorTorque

def get_joint_positions(body, joints): # joints=None):
    return tuple(get_joint_position(body, joint) for joint in joints)

def get_joint_velocities(body, joints):
    return tuple(get_joint_velocity(body, joint) for joint in joints)

def set_joint_position(body, joint, value):
    p.resetJointState(body, joint, value, targetVelocity=0, physicsClientId=CLIENT)

def set_joint_positions(body, joints, values):
    assert len(joints) == len(values), 'joints {} | values {}'.format(joints, values)
    for joint, value in zip(joints, values):
        set_joint_position(body, joint, value)

def get_configuration(body):
    return get_joint_positions(body, get_movable_joints(body))

def set_configuration(body, values):
    set_joint_positions(body, get_movable_joints(body), values)

def get_full_configuration(body):
    # Cannot alter fixed joints
    return get_joint_positions(body, get_joints(body))

def get_labeled_configuration(body):
    movable_joints = get_movable_joints(body)
    return dict(zip(get_joint_names(body, movable_joints),
                    get_joint_positions(body, movable_joints)))

def get_joint_type(body, joint):
    return get_joint_info(body, joint).jointType

def is_fixed(body, joint):
    return get_joint_type(body, joint) == p.JOINT_FIXED

def is_movable(body, joint):
    return not is_fixed(body, joint)

def prune_fixed_joints(body, joints):
    return [joint for joint in joints if is_movable(body, joint)]

def get_movable_joints(body): # 45 / 87 on pr2
    return prune_fixed_joints(body, get_joints(body))

def joint_from_movable(body, index):
    return get_joints(body)[index]

def movable_from_joints(body, joints):
    movable_from_original = {o: m for m, o in enumerate(get_movable_joints(body))}
    return [movable_from_original[joint] for joint in joints]

def is_circular(body, joint):
    joint_info = get_joint_info(body, joint)
    if joint_info.jointType == p.JOINT_FIXED:
        return False
    return joint_info.jointUpperLimit < joint_info.jointLowerLimit

def get_joint_limits(body, joint):
    # TODO: make a version for several joints?
    if is_circular(body, joint):
        # TODO: return UNBOUNDED_LIMITS
        return CIRCULAR_LIMITS
    joint_info = get_joint_info(body, joint)
    return joint_info.jointLowerLimit, joint_info.jointUpperLimit

def get_min_limit(body, joint):
    # TODO: rename to min_position
    return get_joint_limits(body, joint)[0]

def get_min_limits(body, joints):
    return [get_min_limit(body, joint) for joint in joints]

def get_max_limit(body, joint):
    return get_joint_limits(body, joint)[1]

def get_max_limits(body, joints):
    return [get_max_limit(body, joint) for joint in joints]

def get_max_velocity(body, joint):
    return get_joint_info(body, joint).jointMaxVelocity

def get_max_force(body, joint):
    return get_joint_info(body, joint).jointMaxForce

def get_joint_q_index(body, joint):
    return get_joint_info(body, joint).qIndex

def get_joint_v_index(body, joint):
    return get_joint_info(body, joint).uIndex

def get_joint_axis(body, joint):
    return get_joint_info(body, joint).jointAxis

def get_joint_parent_frame(body, joint):
    joint_info = get_joint_info(body, joint)
    return joint_info.parentFramePos, joint_info.parentFrameOrn

def violates_limit(body, joint, value):
    if is_circular(body, joint):
        return False
    lower, upper = get_joint_limits(body, joint)
    return (value < lower) or (upper < value)

def violates_limits(body, joints, values):
    return any(violates_limit(body, joint, value) for joint, value in zip(joints, values))

def wrap_position(body, joint, position):
    from pybullet_planning.interfaces.env_manager.pose_transformation import wrap_angle
    if is_circular(body, joint):
        return wrap_angle(position)
    return position

def wrap_positions(body, joints, positions):
    assert len(joints) == len(positions)
    return [wrap_position(body, joint, position)
            for joint, position in zip(joints, positions)]

def get_custom_limits(body, joints, custom_limits={}, circular_limits=UNBOUNDED_LIMITS):
    joint_limits = []
    for joint in joints:
        if joint in custom_limits:
            joint_limits.append(custom_limits[joint])
        elif is_circular(body, joint):
            joint_limits.append(circular_limits)
        else:
            joint_limits.append(get_joint_limits(body, joint))
    return zip(*joint_limits)

def get_custom_max_velocity(body, joints, custom_vel_limits={}):
    vel_limits = []
    for joint in joints:
        if joint in custom_vel_limits:
            vel_limits.append(custom_vel_limits[joint])
        else:
            vel_limits.append(get_max_velocity(body, joint))
    return vel_limits



from pybullet_planning.interfaces import get_relative_pose, get_link_subtree, clone_body, set_static, get_link_pose, \
    set_pose, multiply, get_pose, invert

##############################################

class EndEffector(object):
    """a convenient class for creating and manipulating an end effector

    Note: the end effector needs to be modeled in the robot's URDF.

    """
    def __init__(self, robot, ee_link, tool_link, **kwargs):
        """[summary]

        Parameters
        ----------
        robot : [type]
            [description]
        ee_link : int
            pb link index of the link where the end effector gets attached to
        tool_link : int
            pb link index of the TCP (tip) link
        """
        self.robot = robot
        self.ee_link = ee_link
        self.tool_link = tool_link
        self.tool_from_ee = get_relative_pose(self.robot, self.ee_link, self.tool_link)
        tool_links = get_link_subtree(robot, self.ee_link)
        self.body = clone_body(robot, links=tool_links, **kwargs)
        set_static(self.body)
        # for link in get_all_links(tool_body):
        #    set_color(tool_body, np.zeros(4), link)
    def get_tool_pose(self, get_cloned_pose=True):
        """[summary]

        Parameters
        ----------
        get_cloned_pose : bool, optional
            if True return the cloned floating EE's tool pose, the robot's tool pose otherwise, by default True

        Returns
        -------
        [type]
            [description]
        """
        if not clone_body:
            return get_link_pose(self.robot, self.tool_link)
        else:
            ee_pose = get_pose(self.body)
            return multiply(ee_pose, invert(self.tool_from_ee))

    def set_pose(self, tool_pose):
        pose = multiply(tool_pose, self.tool_from_ee)
        set_pose(self.body, pose)
        return pose

    @property
    def tool_from_root(self):
        return self.tool_from_ee

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__, self.robot, self.body)



"""
********************************************************************************
interfaces.kinematics
********************************************************************************

.. currentmodule:: pybullet_planning.interfaces.kinematics

Kinematics interface
--------------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    get_ik_tool_link_pose
    get_ik_generator
    sample_tool_ik
    compute_forward_kinematics
    compute_inverse_kinematics
    select_solution

"""


from .ik_utils import *
from .ik_interface import *
from .reachability import *

__all__ = [name for name in dir() if not name.startswith('_')]



import math
import random
import numpy as np

from pybullet_planning.interfaces import Pose, Point, Euler, unit_pose, point_from_pose, multiply, quat_from_euler, invert
from pybullet_planning.interfaces import approximate_as_prism, approximate_as_cylinder


def get_top_grasps(body, under=False):
    raise NotImplementedError()


def get_side_grasps(body, under=False):
    raise NotImplementedError()


def get_side_cylinder_grasps(body, tool_pose=unit_pose(), body_pose=unit_pose(), reverse_grasp=False,
    safety_margin_length=0.0):
    """generate side grasp using the body's bounding cylinder.

    The generated grasp should be used as: ``world_from_body = world_from_gripper * grasp``

    The image below shows the default gripper axis convention.

    .. image:: ../images/cylinder_grasp_convention.png
        :align: center

    Parameters
    ----------
    body : int
        body id for the body being grasped
    tool_pose : Pose, optional
        additional default_ee_from_customized_ee, by default unit_pose()
    body_pose : Pose, optional
        the body's current pose, used for computing bounding box, by default unit_pose()
    reverse_grasp : bool, optional
        generate flipped gripper for pi-symmetry around the gripper's z axis, by default False
    safety_margin_length : float, optional
        safety margin along the gripper's y axis, i.e. the ``top_offset`` shown in the picture, by default 0.0

    Yields
    -------
    Pose
        gripper_from_body
    """
    center, (_, height) = approximate_as_cylinder(body, body_pose=body_pose)
    assert safety_margin_length <= height/2, \
    'safety margin length must be smaller than half of the bounding cylinder\'s height {}'.format(height/2)
    object_from_center = Pose(center - point_from_pose(body_pose))

    # rotate the cylinder's frame to make x axis align with the longitude axis
    longitude_x = Pose(euler=Euler(pitch=np.pi/2))
    while True:
        # rotational symmetry around the longitude axis
        theta = random.uniform(0, 2*np.pi)
        rotate_around_x_axis = Pose(euler=Euler(theta, 0, 0))
        # translation along the longitude axis
        slide_dist = random.uniform(-height/2+safety_margin_length, height/2-safety_margin_length)
        translate_along_x_axis = Pose(point=Point(slide_dist,0,0))

        for j in range(1 + reverse_grasp):
            # the base pi/2 is to make y align with the longitude axis, conforming to the convention (see image in the doc)
            # flip the gripper, gripper symmetry
            rotate_around_z = Pose(euler=[0, 0, math.pi/2 + j * math.pi])

            object_from_gripper = multiply(object_from_center, longitude_x, translate_along_x_axis, rotate_around_x_axis, \
                rotate_around_z, tool_pose)
            yield invert(object_from_gripper)



from scipy.spatial.kdtree import KDTree
from heapq import heappush, heappop
from collections import namedtuple

from pybullet_planning.utils.iter_utils import get_pairs
from .utils import INF, elapsed_time
from .utils import random_selector, default_selector
from .smoothing import smooth_path

import time
import numpy as np

__all__ = [
    'lazy_prm',
    'replan_loop',
    ]

Node = namedtuple('Node', ['g', 'parent'])
unit_cost_fn = lambda v1, v2: 1.
zero_heuristic_fn = lambda v: 0

def retrace_path(visited, vertex):
    if vertex is None:
        return []
    return retrace_path(visited, visited[vertex].parent) + [vertex]

def dijkstra(start_v, neighbors_fn, cost_fn=unit_cost_fn):
    # Update the heuristic over time
    # TODO: overlap with discrete
    start_g = 0
    visited = {start_v: Node(start_g, None)}
    queue = [(start_g, start_v)]
    while queue:
        current_g, current_v = heappop(queue)
        if visited[current_v].g < current_g:
            continue
        for next_v in neighbors_fn(current_v):
            next_g = current_g + cost_fn(current_v, next_v)
            if (next_v not in visited) or (next_g < visited[next_v].g):
                visited[next_v] = Node(next_g, current_v)
                heappush(queue, (next_g, next_v))
    return visited

def wastar_search(start_v, end_v, neighbors_fn, cost_fn=unit_cost_fn,
                  heuristic_fn=zero_heuristic_fn, w=1, max_cost=INF, max_time=INF):
    """WA* uses evaluation function f(n) = g(n) + W*h(n).
    See: https://www.cs.cmu.edu/~aplatzer/orbital/Orbital-doc/api/orbital/algorithm/template/WAStar.html
    """
    # TODO: lazy wastar to get different paths
    #heuristic_fn = lambda v: cost_fn(v, end_v)
    priority_fn = lambda g, h: g + w*h
    goal_test = lambda v: v == end_v

    start_time = time.time()
    start_g, start_h = 0, heuristic_fn(start_v)
    visited = {start_v: Node(start_g, None)}
    queue = [(priority_fn(start_g, start_h), start_g, start_v)]
    while queue and (elapsed_time(start_time) < max_time):
        _, current_g, current_v = heappop(queue)
        if visited[current_v].g < current_g:
            continue
        if goal_test(current_v):
            return retrace_path(visited, current_v)
        for next_v in neighbors_fn(current_v):
            next_g = current_g + cost_fn(current_v, next_v)
            if (next_v not in visited) or (next_g < visited[next_v].g):
                visited[next_v] = Node(next_g, current_v)
                next_h = heuristic_fn(next_v)
                if priority_fn(next_g, next_h) < max_cost:
                    heappush(queue, (priority_fn(next_g, next_h), next_g, next_v))
    return None

##################################################

def get_embed_fn(weights):
    """get embedding function handle that apply dot(weights, q)
    """
    return lambda q: weights * q

def get_distance_fn(weights, p_norm=2):
    embed_fn = get_embed_fn(weights)
    return lambda q1, q2: np.linalg.norm(embed_fn(q2) - embed_fn(q1), ord=p_norm)

##################################################

def check_vertex(v, samples, colliding_vertices, collision_fn):
    if v not in colliding_vertices:
        colliding_vertices[v] = collision_fn(samples[v])
    return not colliding_vertices[v]

def check_edge(v1, v2, samples, colliding_edges, collision_fn, extend_fn):
    if (v1, v2) not in colliding_edges:
        segment = default_selector(extend_fn(samples[v1], samples[v2]))
        colliding_edges[v1, v2] = any(map(collision_fn, segment))
        colliding_edges[v2, v1] = colliding_edges[v1, v2]
    return not colliding_edges[v1, v2]

def check_path(path, colliding_vertices, colliding_edges, samples, extend_fn, collision_fn):
    for v in random_selector(path):
        if not check_vertex(v, samples, colliding_vertices, collision_fn):
            return False
    for v1, v2 in default_selector(get_pairs(path)):
        if not check_edge(v1, v2, samples, colliding_edges, collision_fn, extend_fn):
            return False
    return True

##################################################

def compute_graph(samples, weights=None, p_norm=2, max_degree=10, max_distance=INF, approximate_eps=0.):
    """Build a graph based on samples. We use ``scipy.spatial.kdtree.KDTree`` for now.

    Parameters
    ----------
    samples : list
        list of sampled configurations
    weights : list, optional
        joint weights, by default None, using np.ones(dimension)
    p_norm : int, optional
        p norm used for computing distance, by default 2
    max_degree : int, optional
        max vertex degree in the graph, by default 10
    max_distance : [type], optional
        [description], by default INF
    approximate_eps : [type], optional
        [description], by default 0.

    Returns
    -------
    vertices : list
        list of graph vertex indices
    edges : list
        list of graph edges (vertex index pairs)
    """
    vertices = list(range(len(samples)))
    edges = set()
    if not vertices:
        return vertices, edges
    if weights is None:
        weights = np.ones(len(samples[0]))
    embed_fn = get_embed_fn(weights)
    embedded = list(map(embed_fn, samples))
    kd_tree = KDTree(embedded)
    for v1 in vertices:
        # TODO: could dynamically compute distances
        distances, neighbors = kd_tree.query(embedded[v1], k=max_degree + 1, eps=approximate_eps,
                                             p=p_norm, distance_upper_bound=max_distance)
        for d, v2 in zip(distances, neighbors):
            if (d < max_distance) and (v1 != v2):
                edges.update([(v1, v2), (v2, v1)])
    # print(time.time() - start_time, len(edges), float(len(edges))/len(samples))
    return vertices, edges

##################################################

def lazy_prm(start, goal, sample_fn, extend_fn, collision_fn, num_samples=100,
             weights=None, p_norm=2, lazy=False, max_cost=INF, max_time=INF, verbose=False, draw_fn=None, **kwargs): #, max_paths=INF):
    """
    :param start: Start configuration - conf
    :param goal: End configuration - conf
    :param sample_fn: Sample function - sample_fn()->conf
    :param extend_fn: Extension function - extend_fn(q1, q2)->[q', ..., q"]
    :param collision_fn: Collision function - collision_fn(q)->bool
    :param max_time: Maximum runtime - float
    :param kwargs: Keyword arguments
    :return: Path [q', ..., q"] or None if unable to find a solution
    """
    # TODO: compute parameters using start, goal, and sample_fn statistics
    # TODO: multi-query motion planning
    start_time = time.time()
    # TODO: can embed pose and/or points on the robot for other distances
    if weights is None:
        weights = np.ones(len(start))
    distance_fn = get_distance_fn(weights, p_norm=p_norm)
    # TODO: can compute cost between waypoints from extend_fn

    samples = []
    while len(samples) < num_samples:
        conf = sample_fn()
        if (distance_fn(start, conf) + distance_fn(conf, goal)) < max_cost:
            samples.append(conf)
    start_index, end_index = 0, 1
    samples[start_index] = start
    samples[end_index] = goal
    cost_fn = lambda v1, v2: distance_fn(samples[v1], samples[v2])

    vertices, edges = compute_graph(samples, p_norm=p_norm, **kwargs)
    neighbors_from_index = {v: set() for v in vertices}
    for v1, v2 in edges:
        neighbors_from_index[v1].add(v2)
    if draw_fn:
        draw_fn(samples, edges)

    colliding_vertices, colliding_edges = {}, {}
    def neighbors_fn(v1):
        for v2 in neighbors_from_index[v1]:
            if not colliding_vertices.get(v2, False) and not colliding_edges.get((v1, v2), False):
                yield v2

    if not lazy:
        for vertex in vertices:
            check_vertex(vertex, samples, colliding_vertices, collision_fn)
        for vertex1, vertex2 in edges:
            check_edge(vertex1, vertex2, samples, colliding_edges, collision_fn, extend_fn)

    visited = dijkstra(end_index, neighbors_fn, cost_fn)
    heuristic_fn = lambda v: visited[v].g if v in visited else INF
    path = None
    while (elapsed_time(start_time) < max_time) and (path is None): # TODO: max_attempts
        # TODO: extra cost to prioritize reusing checked edges
        lazy_path = wastar_search(start_index, end_index, neighbors_fn=neighbors_fn,
                                  cost_fn=cost_fn, heuristic_fn=heuristic_fn,
                                  max_cost=max_cost, max_time=max_time-elapsed_time(start_time))
        if lazy_path is None:
            break
        cost = sum(cost_fn(v1, v2) for v1, v2 in get_pairs(lazy_path))
        if verbose:
            print('Length: {} | Cost: {:.3f} | Vertices: {} | Edges: {} | Time: {:.3f}'.format(
                len(lazy_path), cost, len(colliding_vertices), len(colliding_edges), elapsed_time(start_time)))
        if check_path(lazy_path, colliding_vertices, colliding_edges, samples, extend_fn, collision_fn):
            path = lazy_path

    if path is None:
        return path, edges, colliding_vertices, colliding_edges
    solution = [start]
    for q1, q2 in get_pairs(path):
        solution.extend(extend_fn(samples[q1], samples[q2]))
    return solution, samples, edges, colliding_vertices, colliding_edges

##################################################

def replan_loop(start_conf, end_conf, sample_fn, extend_fn, collision_fn, params_list=[100], smooth=0, **kwargs):
    if collision_fn(start_conf) or collision_fn(end_conf):
        return None
    from .meta import direct_path
    path = direct_path(start_conf, end_conf, extend_fn, collision_fn, **kwargs)
    if path is not None:
        return path
    for num_samples in params_list:
        path, _, _, _ = lazy_prm(start_conf, end_conf, sample_fn, extend_fn, collision_fn,
                        num_samples=num_samples, **kwargs)
        if path is not None:
            return smooth_path(path, extend_fn, collision_fn, max_smooth_iterations=smooth, **kwargs)
    return None



from collections.abc import Mapping

#class StarRoadmap(Mapping, object):
class StarRoadmap(Mapping, object):

    def __init__(self, center, planner):
        self.center = center
        self.planner = planner
        self.roadmap = {}

    """
    def __getitem__(self, q):
        return self.roadmap[q]

    def __len__(self):
        return len(self.roadmap)

    def __iter__(self):
        return iter(self.roadmap)
    """

    def grow(self, goal):
        if goal not in self.roadmap:
            self.roadmap[goal] = self.planner(self.center, goal)
        return self.roadmap[goal]

    def __call__(self, start, goal):
        start_traj = self.grow(start)
        if start_traj is None:
            return None
        goal_traj = self.grow(goal)
        if goal_traj is None:
            return None
        return start_traj.reverse(), goal_traj



from itertools import cycle, islice

def implies(p1, p2):
    return not p1 or p2

def roundrobin(*iterables):
    """roundrobin('ABC', 'D', 'EF') --> A D E B F C

    https://docs.python.org/3.1/library/itertools.html#recipes
    Recipe credited to George Sakkis
    """
    pending = len(iterables)
    nexts = cycle(iter(it).__next__ for it in iterables)
    while pending:
        try:
            for next in nexts:
                yield next()
        except StopIteration:
            pending -= 1
            nexts = cycle(islice(nexts, pending))

def chunks(sequence, n=1):
    """get [i, i+n] sublist of a given sequence
    """
    for i in range(0, len(sequence), n):
        yield sequence[i:i + n]

def safe_zip(sequence1, sequence2):
    """zip with safeguarding on the length
    """
    assert len(sequence1) == len(sequence2)
    return zip(sequence1, sequence2)

def get_pairs(sequence):
    """get a sequece of (seq[i], seq[i+1]), i=0~n-1
    """
    return list(safe_zip(sequence[:-1], sequence[1:]))



"""unified entry API for calling different planners
"""
import time

from .lattice import lattice
from .lazy_prm import lazy_prm
from .prm import prm
from .rrt import rrt
from .rrt_connect import rrt_connect, birrt
from .rrt_star import rrt_star

from .utils import INF
from .smoothing import smooth_path
from .utils import RRT_RESTARTS, RRT_SMOOTHING, INF, irange, elapsed_time, compute_path_cost, default_selector, get_pairs, \
    remove_redundant

def direct_path(start, goal, extend_fn, collision_fn, sweep_collision_fn=None, diagnosis=False, **kwargs):
    """direct linear path connnecting start and goal using the extension fn.

    :param start: Start configuration - conf
    :param goal: End configuration - conf
    :param extend_fn: Extension function - extend_fn(q1, q2)->[q', ..., q"]
    :param collision_fn: Collision function - collision_fn(q)->bool
    :param sweep_collision_fn (Optional): Sweep collision function - collision_fn(q0, q1)->bool
    :return: Path [q', ..., q"] or None if unable to find a solution
    """
    # TODO: version which checks whether the segment is valid
    if collision_fn(start) or collision_fn(goal):
        return None
    path = list(extend_fn(start, goal))
    if any(collision_fn(q, diagnosis=diagnosis) for q in default_selector(path)):
        return None
    if sweep_collision_fn is not None:
        if any(sweep_collision_fn(q0, q1, diagnosis=diagnosis) for q0, q1 in default_selector(get_pairs(path))):
            return None
    return path

def check_direct(start, goal, extend_fn, collision_fn, **kwargs):
    if any(collision_fn(q) for q in [start, goal]):
        return False
    return direct_path(start, goal, extend_fn, collision_fn, **kwargs)

#################################################################

def random_restarts(solve_fn, start, goal, distance_fn, sample_fn, extend_fn, collision_fn,
                    restarts=RRT_RESTARTS, smooth=RRT_SMOOTHING,
                    success_cost=0., max_time=INF, max_solutions=1, verbose=False, **kwargs):
    """Apply random restarts to a given planning algorithm to obtain multiple solutions.

    Parameters
    ----------
    solve_fn : function handle
        motion planner function, e.g. ``birrt``
    start : list
        start conf
    goal : list
        end conf
    distance_fn : function handle
        Distance function - `distance_fn(q1, q2)->float`
        see `pybullet_planning.interfaces.planner_interface.joint_motion_planning.get_difference_fn` for an example
    sample_fn : function handle
        configuration space sampler - `sample_fn()->conf`
        see `pybullet_planning.interfaces.planner_interface.joint_motion_planning.get_sample_fn` for an example
    extend_fn : function handle
        Extension function - `extend_fn(q1, q2)->[q', ..., q"]`
        see `pybullet_planning.interfaces.planner_interface.joint_motion_planning.get_extend_fn` for an example
    collision_fn : function handle
        Collision function - `collision_fn(q)->bool`
        see `pybullet_planning.interfaces.robots.collision.get_collision_fn` for an example
    restarts : int, optional
        number of random restarts, by default RRT_RESTARTS
    smooth : int, optional
        smoothing iterations, by default RRT_SMOOTHING
    success_cost : float, optional
        random restarts will terminate early if a path with cost lower than this number is found, by default 0.
    max_time : float, optional
        max allowed runtime, by default INF
    max_solutions : int, optional
        number of solutions wanted, random restarts will terminate early if more solutions are found, by default 1
    verbose : bool, optional
        print toggle, by default False

    Returns
    -------
    list
        list of paths, [[q', ..., q"], [[q'', ..., q""]]
    """
    start_time = time.time()
    solutions = []
    path = check_direct(start, goal, extend_fn, collision_fn, **kwargs)
    if path is False:
        return None
    if path is not None:
        solutions.append(path)

    for attempt in irange(restarts + 1):
        if (len(solutions) >= max_solutions) or (elapsed_time(start_time) >= max_time):
            break
        attempt_time = (max_time - elapsed_time(start_time))
        path = solve_fn(start, goal, distance_fn, sample_fn, extend_fn, collision_fn,
                        max_time=attempt_time, **kwargs)
        if path is None:
            continue
        path = smooth_path(path, extend_fn, collision_fn, max_smooth_iterations=smooth,
                           max_time=max_time-elapsed_time(start_time), **kwargs)
        solutions.append(path)
        if compute_path_cost(path, distance_fn) < success_cost:
            break
    solutions = sorted(solutions, key=lambda path: compute_path_cost(path, distance_fn))
    if verbose:
        print('Solutions ({}): {} | Time: {:.3f}'.format(len(solutions), [(len(path), round(compute_path_cost(
            path, distance_fn), 3)) for path in solutions], elapsed_time(start_time)))
    return solutions

def solve_and_smooth(solve_fn, q1, q2, distance_fn, sample_fn, extend_fn, collision_fn, **kwargs):
    """plan and smooth without random restarting.
    """
    return random_restarts(solve_fn, q1, q2, distance_fn, sample_fn, extend_fn, collision_fn, restarts=0, **kwargs)

#################################################################

def solve_motion_plan(start, goal, distance_fn, sample_fn, extend_fn, collision_fn, algorithm='birrt',
          max_time=INF, max_iterations=INF, num_samples=100, smooth=None, **kwargs):
    # TODO: allow distance_fn to be skipped
    # TODO: return lambda function
    start_time = time.time()
    path = check_direct(start, goal, extend_fn, collision_fn, **kwargs)
    if path is not None:
        return path
    max_time -= elapsed_time(start_time)
    if algorithm == 'prm':
        path = prm(start, goal, distance_fn, sample_fn, extend_fn, collision_fn,
                   num_samples=num_samples, **kwargs)
    elif algorithm == 'lazy_prm':
        path = lazy_prm(start, goal, sample_fn, extend_fn, collision_fn,
                        num_samples=num_samples, max_time=max_time, **kwargs)[0]
    elif algorithm == 'rrt':
        path = rrt(start, goal, distance_fn, sample_fn, extend_fn, collision_fn,
                   max_iterations=max_iterations, max_time=max_time, **kwargs)
    elif algorithm == 'rrt_connect':
        path = rrt_connect(start, goal, distance_fn, sample_fn, extend_fn, collision_fn,
                           max_iterations=max_iterations, max_time=max_time, **kwargs)
    elif algorithm == 'birrt':
        # TODO: checks the straight-line twice
        path = birrt(start, goal, distance_fn, sample_fn, extend_fn, collision_fn,
                     max_iterations=max_iterations, max_time=max_time, smooth=None, **kwargs) # restarts=2
    elif algorithm == 'rrt_star':
        path = rrt_star(start, goal, distance_fn, sample_fn, extend_fn, collision_fn, radius=1,
                        max_iterations=max_iterations, max_time=max_time, **kwargs)
    elif algorithm == 'lattice':
        path = lattice(start, goal, extend_fn, collision_fn, distance_fn=distance_fn, max_time=INF, **kwargs)
    else:
        raise NotImplementedError(algorithm)
    if path:
        path = remove_redundant(path)
    return smooth_path(path, extend_fn, collision_fn, max_smooth_iterations=smooth, max_time=max_time-elapsed_time(start_time), **kwargs)




import numpy as np
import pybullet as p

from pybullet_planning.utils import CLIENT
from pybullet_planning.interfaces.env_manager.pose_transformation import unit_point
from pybullet_planning.interfaces.robots.joint import get_max_velocity, get_max_force, get_joint_positions, get_movable_joints, \
    movable_from_joints

#####################################
# Control

def control_joint(body, joint, value):
    """[summary]

    Parameters
    ----------
    body : [type]
        [description]
    joint : [type]
        [description]
    value : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    return p.setJointMotorControl2(bodyUniqueId=body,
                                   jointIndex=joint,
                                   controlMode=p.POSITION_CONTROL,
                                   targetPosition=value,
                                   targetVelocity=0.0,
                                   maxVelocity=get_max_velocity(body, joint),
                                   force=get_max_force(body, joint),
                                   physicsClientId=CLIENT)

def control_joints(body, joints, positions):
    """[summary]

    Parameters
    ----------
    body : [type]
        [description]
    joints : [type]
        [description]
    positions : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    # TODO: the whole PR2 seems to jitter
    #kp = 1.0
    #kv = 0.3
    #forces = [get_max_force(body, joint) for joint in joints]
    #forces = [5000]*len(joints)
    #forces = [20000]*len(joints)
    return p.setJointMotorControlArray(body, joints, p.POSITION_CONTROL,
                                       targetPositions=positions,
                                       targetVelocities=[0.0] * len(joints),
                                       physicsClientId=CLIENT) #,
                                        #positionGains=[kp] * len(joints),
                                        #velocityGains=[kv] * len(joints),)
                                        #forces=forces)

def joint_controller(body, joints, target, tolerance=1e-3):
    assert(len(joints) == len(target))
    positions = get_joint_positions(body, joints)
    while not np.allclose(positions, target, atol=tolerance, rtol=0):
        control_joints(body, joints, target)
        yield positions
        positions = get_joint_positions(body, joints)

def joint_controller_hold(body, joints, target, **kwargs):
    """
    Keeps other joints in place
    """
    movable_joints = get_movable_joints(body)
    conf = list(get_joint_positions(body, movable_joints))
    for joint, value in zip(movable_from_joints(body, joints), target):
        conf[joint] = value
    return joint_controller(body, movable_joints, conf, **kwargs)

def joint_controller_hold2(body, joints, positions, velocities=None,
                           tolerance=1e-2 * np.pi, position_gain=0.05, velocity_gain=0.01):
    """
    Keeps other joints in place
    """
    # TODO: velocity_gain causes the PR2 to oscillate
    if velocities is None:
        velocities = [0.] * len(positions)
    movable_joints = get_movable_joints(body)
    target_positions = list(get_joint_positions(body, movable_joints))
    target_velocities = [0.] * len(movable_joints)
    movable_from_original = {o: m for m, o in enumerate(movable_joints)}
    #print(list(positions), list(velocities))
    for joint, position, velocity in zip(joints, positions, velocities):
        target_positions[movable_from_original[joint]] = position
        target_velocities[movable_from_original[joint]] = velocity
    # return joint_controller(body, movable_joints, conf)
    current_conf = get_joint_positions(body, movable_joints)
    #forces = [get_max_force(body, joint) for joint in movable_joints]
    while not np.allclose(current_conf, target_positions, atol=tolerance, rtol=0):
        # TODO: only enforce velocity constraints at end
        p.setJointMotorControlArray(body, movable_joints, p.POSITION_CONTROL,
                                    targetPositions=target_positions,
                                    #targetVelocities=target_velocities,
                                    positionGains=[position_gain] * len(movable_joints),
                                    #velocityGains=[velocity_gain] * len(movable_joints),
                                    #forces=forces,
                                    physicsClientId=CLIENT)
        yield current_conf
        current_conf = get_joint_positions(body, movable_joints)

def trajectory_controller(body, joints, path, **kwargs):
    for target in path:
        for positions in joint_controller(body, joints, target, **kwargs):
            yield positions

def simulate_controller(controller, max_time=np.inf): # Allow option to sleep rather than yield?
    from pybullet_planning.interfaces.env_manager import get_time_step, step_simulation
    sim_dt = get_time_step()
    sim_time = 0.0
    for _ in controller:
        if max_time < sim_time:
            break
        step_simulation()
        sim_time += sim_dt
        yield sim_time

def velocity_control_joints(body, joints, velocities):
    #kv = 0.3
    return p.setJointMotorControlArray(body, joints, p.VELOCITY_CONTROL,
                                       targetVelocities=velocities,
                                       physicsClientId=CLIENT) #,
                                        #velocityGains=[kv] * len(joints),)
                                        #forces=forces)

#####################################

def compute_jacobian(robot, link, positions=None):
    joints = get_movable_joints(robot)
    if positions is None:
        positions = get_joint_positions(robot, joints)
    assert len(joints) == len(positions)
    velocities = [0.0] * len(positions)
    accelerations = [0.0] * len(positions)
    translate, rotate = p.calculateJacobian(robot, link, unit_point(), positions,
                                            velocities, accelerations, physicsClientId=CLIENT)
    #movable_from_joints(robot, joints)
    return list(zip(*translate)), list(zip(*rotate)) # len(joints) x 3


def compute_joint_weights(robot, num=100):
    import time
    from pybullet_planning.interfaces.planner_interface.joint_motion_planning import get_sample_fn
    from pybullet_planning.interfaces.robots.link import get_links
    from pybullet_planning.interfaces.robots.dynamics import get_mass

    # http://openrave.org/docs/0.6.6/_modules/openravepy/databases/linkstatistics/#LinkStatisticsModel
    # TODO: use velocities instead
    start_time = time.time()
    joints = get_movable_joints(robot)
    sample_fn = get_sample_fn(robot, joints)
    weighted_jacobian = np.zeros(len(joints))
    links = list(get_links(robot))
    # links = {l for j in joints for l in get_link_descendants(self.robot, j)}
    masses = [get_mass(robot, link) for link in links]  # Volume, AABB volume
    total_mass = sum(masses)
    for _ in range(num):
        conf = sample_fn()
        for mass, link in zip(masses, links):
            translate, rotate = compute_jacobian(robot, link, conf)
            weighted_jacobian += np.array([mass * np.linalg.norm(vec) for vec in translate]) / total_mass
    weighted_jacobian /= num
    print(list(weighted_jacobian))
    print(time.time() - start_time)
    return weighted_jacobian

#####################################



import numpy as np

from pybullet_planning.utils import MAX_DISTANCE, CIRCULAR_LIMITS
from pybullet_planning.motion_planners import birrt, direct_path

#####################################
# SE(2) pose motion planning

def get_base_difference_fn():
    from pybullet_planning.interfaces.env_manager.pose_transformation import circular_difference
    def fn(q2, q1):
        dx, dy = np.array(q2[:2]) - np.array(q1[:2])
        dtheta = circular_difference(q2[2], q1[2])
        return (dx, dy, dtheta)
    return fn

def get_base_distance_fn(weights=1*np.ones(3)):
    difference_fn = get_base_difference_fn()
    def fn(q1, q2):
        difference = np.array(difference_fn(q2, q1))
        return np.sqrt(np.dot(weights, difference * difference))
    return fn

def plan_base_motion(body, end_conf, base_limits, obstacles=[], direct=False,
                     weights=1*np.ones(3), resolutions=0.05*np.ones(3),
                     max_distance=MAX_DISTANCE, **kwargs):
    from pybullet_planning.interfaces.robots.collision import pairwise_collision
    from pybullet_planning.interfaces.robots.body import set_base_values, get_base_values

    def sample_fn():
        x, y = np.random.uniform(*base_limits)
        theta = np.random.uniform(*CIRCULAR_LIMITS)
        return (x, y, theta)

    difference_fn = get_base_difference_fn()
    distance_fn = get_base_distance_fn(weights=weights)

    def extend_fn(q1, q2):
        steps = np.abs(np.divide(difference_fn(q2, q1), resolutions))
        n = int(np.max(steps)) + 1
        q = q1
        for i in range(n):
            q = tuple((1. / (n - i)) * np.array(difference_fn(q2, q)) + q)
            yield q
            # TODO: should wrap these joints

    def collision_fn(q):
        # TODO: update this function
        set_base_values(body, q)
        return any(pairwise_collision(body, obs, max_distance=max_distance) for obs in obstacles)

    start_conf = get_base_values(body)
    if collision_fn(start_conf):
        print("Warning: initial configuration is in collision")
        return None
    if collision_fn(end_conf):
        print("Warning: end configuration is in collision")
        return None
    if direct:
        return direct_path(start_conf, end_conf, extend_fn, collision_fn)
    return birrt(start_conf, end_conf, distance_fn,
                 sample_fn, extend_fn, collision_fn, **kwargs)



"""
********************************************************************************
interfaces.geometry
********************************************************************************

.. currentmodule:: pybullet_planning.interfaces.geometry

TODO: module description

Main Types
--------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    AABB

Bounding box operations
----------------------------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    aabb_from_points
    aabb_union
    get_bodies_in_region

"""

from __future__ import absolute_import

from .bounding_box import *
from .mesh import *
from .polygon import *
from .pointcloud import *
from .camera import *

__all__ = [name for name in dir() if not name.startswith('_')]



"""
********************************************************************************
utils
********************************************************************************

.. currentmodule:: pybullet_planning.utils

Package containing a set of utility functions and variables

File system functions
=====================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    read_pickle
    write_pickle

Sampling functions
===================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    randomize

Iteration utilities
===================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    roundrobin
    chunks
    get_pairs


Transformation functions
========================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    rotation_from_matrix

"""

from .shared_const import *
from .file_io import *
from .numeric_sample import *
from .transformations import *
from .iter_utils import *
from .debug_utils import *
# from ._file_path_archived import *

__all__ = [name for name in dir() if not name.startswith('_')]



import math
import numpy as np
import pybullet as p

from pybullet_planning.utils import CLIENT

def inverse_kinematics_helper(robot, link, target_pose, null_space=None):
    (target_point, target_quat) = target_pose
    assert target_point is not None
    try:
        if null_space is not None:
            assert target_quat is not None
            lower, upper, ranges, rest = null_space

            kinematic_conf = p.calculateInverseKinematics(robot, link, target_point, target_quat,
                                                          lowerLimits=lower, upperLimits=upper, jointRanges=ranges, restPoses=rest,
                                                          physicsClientId=CLIENT,
                                                        #   maxNumIterations=1000,
                                                        #   residualThreshold=1e-12,
                                                          )
        elif target_quat is None:
            #ikSolver = p.IK_DLS or p.IK_SDLS
            kinematic_conf = p.calculateInverseKinematics(robot, link, target_point,
                                                          #lowerLimits=ll, upperLimits=ul, jointRanges=jr, restPoses=rp, jointDamping=jd,
                                                          # solver=ikSolver, maxNumIterations=-1, residualThreshold=-1,
                                                          physicsClientId=CLIENT)
        else:
            # ! normal case
            kinematic_conf = p.calculateInverseKinematics(robot, link, target_point, target_quat,
                                                          physicsClientId=CLIENT,
                                                        #   maxNumIterations=1000,
                                                        #   residualThreshold=1e-12,
                                                          )
    except p.error as e:
        kinematic_conf = None
    if (kinematic_conf is None) or any(map(math.isnan, kinematic_conf)):
        return None
    return kinematic_conf

def is_pose_close(pose, target_pose, pos_tolerance=1e-3, ori_tolerance=1e-3*np.pi):
    (point, quat) = pose
    (target_point, target_quat) = target_pose
    if (target_point is not None) and not np.allclose(point, target_point, atol=pos_tolerance, rtol=0):
        return False
    if (target_quat is not None) and not (np.allclose(quat, target_quat, atol=ori_tolerance, rtol=0) or \
            np.allclose(quat, -np.array(target_quat), atol=ori_tolerance, rtol=0)):
        return False
    return True

def inverse_kinematics(robot, link, target_pose, max_iterations=200, custom_limits={}, **kwargs):
    from pybullet_planning.interfaces.env_manager.pose_transformation import all_between
    from pybullet_planning.interfaces.robots import get_movable_joints, set_joint_positions, get_link_pose, get_custom_limits

    movable_joints = get_movable_joints(robot)
    for iterations in range(max_iterations):
        # TODO: stop is no progress
        # TODO: stop if collision or invalid joint limits
        kinematic_conf = inverse_kinematics_helper(robot, link, target_pose)
        if kinematic_conf is None:
            return None
        set_joint_positions(robot, movable_joints, kinematic_conf)
        if is_pose_close(get_link_pose(robot, link), target_pose, **kwargs):
            break
    else:
        return None
    lower_limits, upper_limits = get_custom_limits(robot, movable_joints, custom_limits)
    if not all_between(lower_limits, kinematic_conf, upper_limits):
        return None
    return kinematic_conf


def snap_sols(sols, q_guess, joint_limits, weights=None, best_sol_only=False):
    """get the best solution based on closeness to the q_guess and weighted joint diff

    Parameters
    ----------
    sols : [type]
        [description]
    q_guess : [type]
        [description]
    joint_limits : [type]
        [description]
    weights : [type], optional
        [description], by default None
    best_sol_only : bool, optional
        [description], by default False

    Returns
    -------
    lists of joint conf (list)
    or
    joint conf (list)
    """
    import numpy as np
    valid_sols = []
    dof = len(q_guess)
    if not weights:
        weights = [1.0] * dof
    else:
        assert dof == len(weights)

    for sol in sols:
        test_sol = np.ones(dof)*9999.
        for i in range(dof):
            for add_ang in [-2.*np.pi, 0, 2.*np.pi]:
                test_ang = sol[i] + add_ang
                if (test_ang <= joint_limits[i][1] and test_ang >= joint_limits[i][0] and \
                    abs(test_ang - q_guess[i]) < abs(test_sol[i] - q_guess[i])):
                    test_sol[i] = test_ang
        if np.all(test_sol != 9999.):
            valid_sols.append(test_sol.tolist())

    if len(valid_sols) == 0:
        return []

    if best_sol_only:
        best_sol_ind = np.argmin(np.sum((weights*(valid_sols - np.array(q_guess)))**2,1))
        return valid_sols[best_sol_ind]
    else:
        return valid_sols



"""
********************************************************************************
primitives
********************************************************************************

These modules are transported from `kuka_primitives`, `pr2_primitives` and `pr2_utils` from
`ss-pybullet <https://github.com/caelan/ss-pybullet/tree/master/pybullet_tools>`_.

.. currentmodule:: pybullet_planning.primitives

Grasp
------

In ``pybullet_planning``, a grasp is modeled as an operator on the object frame, i.e.
``world_from_object = world_from_gripper * grasp``, where ``grasp = gripper_from_object``.

We provide some simple grasp generator based on bounding geometry (box, cylinder). Inside these
generators' implementation, however, the modeling start from the object frame to the gripper's
frame to obtain ``object_from_gripper`` and then we return its inverse ``gripper_from_object``.
(Yijiang thinks this is more straightforward to think about.)

.. currentmodule:: pybullet_planning.primitives.grasp_gen

.. autosummary::
    :toctree: generated/
    :nosignatures:

    get_side_cylinder_grasps

Convenient classes
------------------

.. currentmodule:: pybullet_planning.primitives.trajectory

.. autosummary::
    :toctree: generated/
    :nosignatures:

    EndEffector

"""

from __future__ import absolute_import

from .grasp_gen import *
from .trajectory import *

__all__ = [name for name in dir() if not name.startswith('_')]





def read_pcd_file(path):
    """
    Reads a *.pcd pointcloud file
    :param path: path to the *.pcd pointcloud file
    :return: list of points
    """
    with open(path) as f:
        data = f.readline().split()
        num_points = 0
        while data[0] != 'DATA':
            if data[0] == 'POINTS':
                num_points = int(data[1])
            data = f.readline().split()
            continue
        return [tuple(map(float, f.readline().split())) for _ in range(num_points)]



from itertools import product, combinations
from collections import defaultdict, deque, namedtuple
import pybullet as p

from pybullet_planning.utils import BASE_LINK, CLIENT
from pybullet_planning.interfaces.robots.joint import get_num_joints, get_joints, get_joint_info, is_movable, prune_fixed_joints

#####################################
# Links

get_num_links = get_num_joints
get_links = get_joints # Does not include BASE_LINK

def child_link_from_joint(joint):
    return joint # link

def parent_joint_from_link(link):
    return link # joint

def get_all_links(body):
    return [BASE_LINK] + list(get_links(body))

def get_link_name(body, link):
    from pybullet_planning.interfaces.robots.body import get_base_name
    if link == BASE_LINK:
        return get_base_name(body)
    return get_joint_info(body, link).linkName.decode('UTF-8')

def get_link_parent(body, link):
    if link == BASE_LINK:
        return None
    return get_joint_info(body, link).parentIndex

parent_link_from_joint = get_link_parent

def link_from_name(body, name):
    from pybullet_planning.interfaces.robots.body import get_base_name
    if name == get_base_name(body):
        return BASE_LINK
    for link in get_joints(body):
        if get_link_name(body, link) == name:
            return link
    raise ValueError(body, name)


def has_link(body, name):
    try:
        link_from_name(body, name)
    except ValueError:
        return False
    return True

LinkState = namedtuple('LinkState', ['linkWorldPosition', 'linkWorldOrientation',
                                     'localInertialFramePosition', 'localInertialFrameOrientation',
                                     'worldLinkFramePosition', 'worldLinkFrameOrientation'])

def get_link_state(body, link, kinematics=True, velocity=True):
    # TODO: the defaults are set to False?
    # https://github.com/bulletphysics/bullet3/blob/master/examples/pybullet/pybullet.c
    return LinkState(*p.getLinkState(body, link, #computeLinkVelocity=velocity, computeForwardKinematics=kinematics,
                                     physicsClientId=CLIENT))

def get_com_pose(body, link): # COM = center of mass
    link_state = get_link_state(body, link)
    return link_state.linkWorldPosition, link_state.linkWorldOrientation

def get_link_inertial_pose(body, link):
    link_state = get_link_state(body, link)
    return link_state.localInertialFramePosition, link_state.localInertialFrameOrientation

def get_link_pose(body, link):
    from pybullet_planning.interfaces.env_manager.pose_transformation import get_pose
    if link == BASE_LINK:
        return get_pose(body)
    # if set to 1 (or True), the Cartesian world position/orientation will be recomputed using forward kinematics.
    link_state = get_link_state(body, link) #, kinematics=True, velocity=False)
    return link_state.worldLinkFramePosition, link_state.worldLinkFrameOrientation

def get_relative_pose(body, link1, link2=BASE_LINK):
    """compute relative transformation between two links of a body

    Parameters
    ----------
    body : [type]
        [description]
    link1 : [type]
        [description]
    link2 : [type], optional
        [description], by default BASE_LINK

    Returns
    -------
    [type]
        [description]
    """
    from pybullet_planning.interfaces.env_manager.pose_transformation import multiply, invert
    world_from_link1 = get_link_pose(body, link1)
    world_from_link2 = get_link_pose(body, link2)
    link2_from_link1 = multiply(invert(world_from_link2), world_from_link1)
    return link2_from_link1

#####################################

def get_all_link_parents(body):
    return {link: get_link_parent(body, link) for link in get_links(body)}

def get_all_link_children(body):
    children = {}
    for child, parent in get_all_link_parents(body).items():
        if parent not in children:
            children[parent] = []
        children[parent].append(child)
    return children

def get_link_children(body, link):
    children = get_all_link_children(body)
    return children.get(link, [])

def get_link_ancestors(body, link):
    # Returns in order of depth
    # Does not include link
    parent = get_link_parent(body, link)
    if parent is None:
        return []
    return get_link_ancestors(body, parent) + [parent]

def get_ordered_ancestors(robot, link):
    #return prune_fixed_joints(robot, get_link_ancestors(robot, link)[1:] + [link])
    return get_link_ancestors(robot, link)[1:] + [link]

def get_joint_ancestors(body, joint):
    link = child_link_from_joint(joint)
    return get_link_ancestors(body, link) + [link]

def get_movable_joint_ancestors(body, link):
    return prune_fixed_joints(body, get_joint_ancestors(body, link))

def get_joint_descendants(body, link):
    return list(map(parent_joint_from_link, get_link_descendants(body, link)))

def get_movable_joint_descendants(body, link):
    return prune_fixed_joints(body, get_joint_descendants(body, link))

def get_link_descendants(body, link, test=lambda l: True):
    descendants = []
    for child in get_link_children(body, link):
        if test(child):
            descendants.append(child)
            descendants.extend(get_link_descendants(body, child, test=test))
    return descendants

def get_link_subtree(body, link, **kwargs):
    return [link] + get_link_descendants(body, link, **kwargs)

def are_links_adjacent(body, link1, link2):
    return (get_link_parent(body, link1) == link2) or \
           (get_link_parent(body, link2) == link1)

def get_adjacent_links(body):
    adjacent = set()
    for link in get_links(body):
        parent = get_link_parent(body, link)
        adjacent.add((link, parent))
        #adjacent.add((parent, link))
    return adjacent

def get_adjacent_fixed_links(body):
    return list(filter(lambda item: not is_movable(body, item[0]),
                       get_adjacent_links(body)))

def get_fixed_links(body):
    edges = defaultdict(list)
    for link, parent in get_adjacent_fixed_links(body):
        edges[link].append(parent)
        edges[parent].append(link)
    visited = set()
    fixed = set()
    for initial_link in get_links(body):
        if initial_link in visited:
            continue
        cluster = [initial_link]
        queue = deque([initial_link])
        visited.add(initial_link)
        while queue:
            for next_link in edges[queue.popleft()]:
                if next_link not in visited:
                    cluster.append(next_link)
                    queue.append(next_link)
                    visited.add(next_link)
        fixed.update(product(cluster, cluster))
    return fixed


def get_moving_links(body, joints):
    moving_links = set()
    for joint in joints:
        link = child_link_from_joint(joint)
        if link not in moving_links:
            moving_links.update(get_link_subtree(body, link))
    return list(moving_links)


def get_moving_pairs(body, moving_joints):
    """
    Check all fixed and moving pairs
    Do not check all fixed and fixed pairs
    Check all moving pairs with a common
    """
    moving_links = get_moving_links(body, moving_joints)
    for link1, link2 in combinations(moving_links, 2):
        ancestors1 = set(get_joint_ancestors(body, link1)) & set(moving_joints)
        ancestors2 = set(get_joint_ancestors(body, link2)) & set(moving_joints)
        if ancestors1 != ancestors2:
            yield link1, link2


def get_self_link_pairs(body, joints, disabled_collisions=set(), only_moving=True):
    moving_links = get_moving_links(body, joints)
    fixed_links = list(set(get_links(body)) - set(moving_links))
    check_link_pairs = list(product(moving_links, fixed_links))
    if only_moving:
        check_link_pairs.extend(get_moving_pairs(body, joints))
    else:
        check_link_pairs.extend(combinations(moving_links, 2))
    check_link_pairs = list(filter(lambda pair: not are_links_adjacent(body, *pair), check_link_pairs))
    check_link_pairs = list(filter(lambda pair: (pair not in disabled_collisions) and
                                                (pair[::-1] not in disabled_collisions), check_link_pairs))
    return check_link_pairs


def get_link_attached_body_pairs(body, attachments=[]):
    link_body_pairs = []
    body_links = get_links(body)

    for attach in attachments:
        if attach.child in [pair[1] for pair in link_body_pairs]:
            continue
        if attach.parent == body:
            body_check_links = list(filter(lambda b_l : attach.parent_link != b_l and \
                not are_links_adjacent(body, b_l, attach.parent_link), body_links))
            link_body_pairs.append((body_check_links, attach.child))

    return link_body_pairs


def get_disabled_collisions(robot, disabled_link_pair_names):
    return {tuple(link_from_name(robot, link)
                  for link in pair if has_link(robot, link))
                  for pair in disabled_link_pair_names}


def get_body_body_disabled_collisions(body1, body2, disabled_link_pair_names):
    disabled_link_pairs = set()
    for link_name1, link_name2 in disabled_link_pair_names:
        if has_link(body1, link_name1) and has_link(body2, link_name2):
            pair = ((body1, link_from_name(body1, link_name1)), (body2, link_from_name(body2, link_name2)))
        elif has_link(body2, link_name1) and has_link(body1, link_name2):
            pair = ((body1, link_from_name(body1, link_name2)), (body2, link_from_name(body2, link_name1)))
        else:
            continue
        disabled_link_pairs.add(pair)
    return disabled_link_pairs



import math
import numpy as np
import pybullet as p
from itertools import product, combinations

from pybullet_planning.utils import CLIENT, BASE_LINK, GREEN, RED, BLUE, BLACK, WHITE, NULL_ID, YELLOW, LOGGER, GREY

from pybullet_planning.interfaces.env_manager.pose_transformation import unit_pose, tform_point, unit_from_theta, get_distance
from pybullet_planning.interfaces.geometry.bounding_box import get_aabb
from pybullet_planning.interfaces.geometry.camera import apply_alpha, set_camera_pose

def get_lifetime(lifetime):
    if lifetime is None:
        return 0
    return lifetime

def add_debug_parameter():
    # TODO: make a slider that controls the step in the trajectory
    # TODO: could store a list of savers
    #targetVelocitySlider = p.addUserDebugParameter("wheelVelocity", -10, 10, 0)
    #maxForce = p.readUserDebugParameter(maxForceSlider)
    raise NotImplementedError()

def add_text(text, position=(0, 0, 0), color=BLACK, lifetime=None, parent=NULL_ID, parent_link=BASE_LINK):
    return p.addUserDebugText(str(text), textPosition=position, textColorRGB=color[:3], # textSize=1,
                              lifeTime=get_lifetime(lifetime), parentObjectUniqueId=parent, parentLinkIndex=parent_link,
                              physicsClientId=CLIENT)

def add_line(start, end, color=BLACK, width=1, lifetime=None, parent=NULL_ID, parent_link=BASE_LINK):
    """[summary]

    Parameters
    ----------
    start : [type]
        [description]
    end : [type]
        [description]
    color : tuple, optional
        [description], by default (0, 0, 0)
    width : int, optional
        [description], by default 1
    lifetime : [type], optional
        [description], by default None
    parent : int, optional
        [description], by default NULL_ID
    parent_link : [type], optional
        [description], by default BASE_LINK

    Returns
    -------
    [type]
        [description]
    """
    return p.addUserDebugLine(start, end, lineColorRGB=color[:3], lineWidth=width,
                              lifeTime=get_lifetime(lifetime), parentObjectUniqueId=parent, parentLinkIndex=parent_link,
                              physicsClientId=CLIENT)

def remove_debug(debug):
    p.removeUserDebugItem(debug, physicsClientId=CLIENT)

remove_handle = remove_debug

def remove_handles(handles):
    for handle in handles:
        remove_debug(handle)

def remove_all_debug():
    p.removeAllUserDebugItems(physicsClientId=CLIENT)

def add_body_name(body, name=None, **kwargs):
    from pybullet_planning.interfaces.env_manager.pose_transformation import set_pose
    from pybullet_planning.interfaces.env_manager.savers import PoseSaver
    from pybullet_planning.interfaces.robots.body import get_name

    name = name or get_name(body)
    with PoseSaver(body):
        set_pose(body, unit_pose())
        lower, upper = get_aabb(body)
    #position = (0, 0, upper[2])
    position = upper
    return add_text(name, position=position, parent=body, **kwargs)  # removeUserDebugItem

def add_segments(points, closed=False, **kwargs):
    lines = []
    for v1, v2 in zip(points, points[1:]):
        lines.append(add_line(v1, v2, **kwargs))
    if closed:
        lines.append(add_line(points[-1], points[0], **kwargs))
    return lines

def draw_link_name(body, link=BASE_LINK, name=None):
    from pybullet_planning.interfaces.robots.link import get_link_name
    return add_text(name or get_link_name(body, link), position=(0, 0.2, 0),
                    parent=body, parent_link=link)

def draw_pose(pose, length=0.1, **kwargs):
    origin_world = tform_point(pose, np.zeros(3))
    handles = []
    for k in range(3):
        axis = np.zeros(3)
        axis[k] = 1
        axis_world = tform_point(pose, length*axis)
        handles.append(add_line(origin_world, axis_world, color=axis, **kwargs))
    return handles

def draw_base_limits(limits, z=1e-2, **kwargs):
    lower, upper = limits
    vertices = [(lower[0], lower[1], z), (lower[0], upper[1], z),
                (upper[0], upper[1], z), (upper[0], lower[1], z)]
    return add_segments(vertices, closed=True, **kwargs)

def draw_circle(center, radius, n=24, **kwargs):
    vertices = []
    for i in range(n):
        theta = i*2*math.pi/n
        unit = np.append(unit_from_theta(theta), [0])
        vertices.append(center+radius*unit)
    return add_segments(vertices, closed=True, **kwargs)

def draw_aabb(aabb, **kwargs):
    d = len(aabb[0])
    vertices = list(product(range(len(aabb)), repeat=d))
    lines = []
    for i1, i2 in combinations(vertices, 2):
        if sum(i1[k] != i2[k] for k in range(d)) == 1:
            p1 = [aabb[i1[k]][k] for k in range(d)]
            p2 = [aabb[i2[k]][k] for k in range(d)]
            lines.append(add_line(p1, p2, **kwargs))
    return lines

def draw_point(point, size=0.01, **kwargs):
    lines = []
    for i in range(len(point)):
        axis = np.zeros(len(point))
        axis[i] = 1.0
        p1 = np.array(point) - size/2 * axis
        p2 = np.array(point) + size/2 * axis
        lines.append(add_line(p1, p2, **kwargs))
    return lines
    #extent = size * np.ones(len(point)) / 2
    #aabb = np.array(point) - extent, np.array(point) + extent
    #return draw_aabb(aabb, **kwargs)

def get_face_edges(face):
    #return list(combinations(face, 2))
    return list(zip(face, face[1:] + face[:1]))

def draw_mesh(mesh, **kwargs):
    verts, faces = mesh
    lines = []
    for face in faces:
        for i1, i2 in get_face_edges(face):
            lines.append(add_line(verts[i1], verts[i2], **kwargs))
    return lines

def draw_ray(ray, ray_result=None, visible_color=GREEN, occluded_color=RED, **kwargs):
    if ray_result is None:
        return [add_line(ray.start, ray.end, color=visible_color, **kwargs)]
    if ray_result.objectUniqueId == NULL_ID:
        hit_position = ray.end
    else:
        hit_position = ray_result.hit_position
    return [
        add_line(ray.start, hit_position, color=visible_color, **kwargs),
        add_line(hit_position, ray.end, color=occluded_color, **kwargs),
    ]


def draw_collision_diagnosis(pb_closest_pt_output, viz_last_duration=-1, point_color=BLACK, line_color=YELLOW, \
    focus_camera=True, camera_ray=np.array([0.1, 0, 0.05]), body_name_from_id=None, viz_all=False, distance_threshold=0.0, max_distance=0.0):
    from pybullet_planning.interfaces.env_manager.simulation import has_gui
    from pybullet_planning.interfaces.env_manager.user_io import wait_for_user, wait_for_duration
    from pybullet_planning.interfaces.robots.link import get_link_name
    from pybullet_planning.interfaces.robots.body import set_color, get_name, clone_body, remove_body, set_pose
    if not pb_closest_pt_output:
        return

    body_name_from_id = body_name_from_id or {}
    for u_cr in pb_closest_pt_output:
        pen_dist = get_distance(u_cr[5], u_cr[6])
        if pen_dist < distance_threshold:
            continue

        handles = []
        b1 = u_cr[1]
        b2 = u_cr[2]
        l1 = u_cr[3]
        l2 = u_cr[4]
        b1_name = body_name_from_id[b1] if b1 in body_name_from_id else get_name(b1)
        b2_name = body_name_from_id[b2] if b2 in body_name_from_id else get_name(b2)
        l1_name = get_link_name(b1, l1)
        l2_name = get_link_name(b2, l2)

        LOGGER.warning('pairwise link collision: (Body #{0}, Link #{1}) - (Body #{2}, Link #{3})'.format(
            b1_name, l1_name, b2_name, l2_name))
        LOGGER.warning('Penetration depth: {:.6f} (m) | point1 ({:.6f},{:.6f},{:.6f}), point2 ({:.6f},{:.6f},{:.6f})'.format(
            pen_dist, *u_cr[5], *u_cr[6]))

        if has_gui():
            # * to not obscure the cloned body color
            set_color(b1, apply_alpha(WHITE, 0.2), link=l1)
            set_color(b2, apply_alpha(WHITE, 0.2), link=l2)

            cloned_b1 = clone_body(b1, links=[l1], collision=True, visual=False)
            cloned_b2 = clone_body(b2, links=[l2], collision=True, visual=False)
            set_color(cloned_b1, apply_alpha(RED, 0.2))
            set_color(cloned_b2, apply_alpha(GREEN, 0.2))

            handles.append(add_body_name(b1, name=b1_name))
            handles.append(add_body_name(b2, name=b2_name))
            handles.append(draw_link_name(b1, l1, name=l1_name))
            handles.append(draw_link_name(b2, l2, name=l2_name))

            handles.append(add_line(u_cr[5], u_cr[6], color=line_color, width=5))
            handles.extend(draw_point(u_cr[5], size=0.002, color=point_color))
            handles.extend(draw_point(u_cr[6], size=0.002, color=point_color))
            if focus_camera:
                camera_base_pt = u_cr[5]
                camera_pt = np.array(camera_base_pt) + camera_ray
                set_camera_pose(tuple(camera_pt), camera_base_pt)

            if viz_last_duration < 0:
                wait_for_user('Visualize collision. Press Enter to continue.')
            else:
                wait_for_duration(viz_last_duration)

            # * restore lines and colors
            remove_handles(handles)
            remove_body(cloned_b1)
            remove_body(cloned_b2)
            # TODO cannot retrieve original color yet
            set_color(b1, apply_alpha(GREY, 1), link=l1)
            set_color(b2, apply_alpha(GREY, 1), link=l2)
        # else:
        #     wait_for_user('Collision diagnosis. Press Enter to continue.')

        if not viz_all:
            return

def draw_ray_result_diagnosis(ray, ray_result, sweep_body=None, sweep_link=None,
        point_color=BLACK, focus_camera=True, camera_ray=np.array([0.1, 0, 0.05]), body_name_from_id=None):
    from pybullet_planning.interfaces.env_manager.simulation import has_gui
    from pybullet_planning.interfaces.env_manager.user_io import wait_for_user
    from pybullet_planning.interfaces.robots.link import get_link_name
    from pybullet_planning.interfaces.robots.body import set_color, get_name, clone_body, remove_body
    if not ray_result:
        return
    body_name_from_id = body_name_from_id or {}

    handles = []
    sweep_body_name = body_name_from_id[sweep_body] if sweep_body is not None and sweep_body in body_name_from_id else None
    sweep_link_name = get_link_name(sweep_body, sweep_link) if sweep_link is not None and sweep_body is not None else None

    b2 = ray_result.objectUniqueId
    l2 = ray_result.linkIndex
    b2_name = body_name_from_id[b2] if b2 in body_name_from_id else get_name(b2)
    l2_name = get_link_name(b2, l2)

    LOGGER.warning('ray collision: Sweeping (Body #{0}, Link #{1}) -> (Body #{2}, Link #{3})'.format(
        sweep_body_name, sweep_link_name, b2_name, l2_name))
    LOGGER.warning('hit_fraction: {:.2f} | hit_position ({:.6f},{:.6f},{:.6f}) | hit_normal ({:.6f},{:.6f},{:.6f})'.format(
        ray_result.hit_fraction, *ray_result.hit_position, *ray_result.hit_normal))

    if has_gui():
        # to not obscure cloned body color
        set_color(b2, apply_alpha(WHITE, 0.5), link=l2)

        cloned_b2 = clone_body(b2, links=[l2], collision=True, visual=False)
        set_color(cloned_b2, apply_alpha(GREEN, 0.2))

        if sweep_body is not None and sweep_link is not None:
            # to not obscure the cloned body color
            set_color(sweep_body, apply_alpha(WHITE, 0.5), link=sweep_link)

            cloned_sweep_body = clone_body(sweep_body, links=[sweep_link], collision=True, visual=False)
            set_color(cloned_sweep_body, apply_alpha(RED, 0.2))
            handles.append(add_body_name(sweep_body, name=sweep_body_name))
            handles.append(draw_link_name(sweep_body, sweep_link, name=sweep_link_name))

        handles.append(add_body_name(b2, name=b2_name))
        handles.append(draw_link_name(b2, l2, name=l2_name))

        # handles.append(add_line(u_cr[5], u_cr[6], color=line_color, width=5))
        handles.extend(draw_ray(ray, ray_result, width=3.0))
        handles.extend(draw_point(ray_result.hit_position, size=0.002, color=point_color))

        if focus_camera:
            camera_base_pt = ray_result.hit_position
            camera_pt = np.array(camera_base_pt) + camera_ray
            set_camera_pose(tuple(camera_pt), camera_base_pt)

        wait_for_user('Visualize collision. Press Enter to continue.')

        # * restore lines and colors
        remove_handles(handles)
        if sweep_body is not None and sweep_link is not None:
            remove_body(cloned_sweep_body)
            set_color(sweep_body, apply_alpha(GREY, 1), link=sweep_link)
        remove_body(cloned_b2)
        # # TODO cannot retrieve original color yet
        set_color(b2, apply_alpha(GREY, 1), link=l2)

    # else:
    #     wait_for_user('Ray collision diagnosis. Press Enter to continue.')


def camera_focus_on_body(body, camera_ray=np.array([0.1, 0, 0.05])):
    from pybullet_planning.interfaces.env_manager.pose_transformation import get_pose
    camera_base_pt, _ = get_pose(body)
    camera_pt = np.array(camera_base_pt) + camera_ray
    set_camera_pose(tuple(camera_pt), camera_base_pt)


def camera_focus_on_point(point, camera_ray=np.array([0.1, 0, 0.05])):
    camera_pt = np.array(point) + camera_ray
    set_camera_pose(tuple(camera_pt), point)



import os
import sys
import time
import numpy as np
import pybullet as p
from collections import namedtuple

from pybullet_planning.utils import INF, CLIENT, CLIENTS
from pybullet_planning.utils import is_darwin

# from future_builtins import map, filter
# from builtins import input # TODO - use future
try:
    user_input = raw_input
except NameError:
    user_input = input

#####################################

# https://stackoverflow.com/questions/5081657/how-do-i-prevent-a-c-shared-library-to-print-on-stdout-in-python/14797594#14797594
# https://stackoverflow.com/questions/4178614/suppressing-output-of-module-calling-outside-library
# https://stackoverflow.com/questions/4675728/redirect-stdout-to-a-file-in-python/22434262#22434262

class HideOutput(object):
    '''
    A context manager that block stdout for its scope, usage:
    with HideOutput():
        os.system('ls -l')
    '''
    DEFAULT_ENABLE = True
    def __init__(self, enable=None):
        if enable is None:
            enable = self.DEFAULT_ENABLE

        # # ! Pytest interferes with file descriptor capture.
        # called_from_test = 'pytest' in sys.modules
        # # ! jupyter notebook (ipython) raises UnsupportedOperation on fileno(): https://github.com/ipython/ipython/pull/3072/
        # called_from_ipykernel = 'ipykernel' in sys.modules
        # enable = False if called_from_test or called_from_ipykernel else enable

        self.enable = enable
        if not self.enable:
            return
        try:
            sys.stdout.flush()
            self._origstdout = sys.stdout
            self._oldstdout_fno = os.dup(sys.stdout.fileno())
            self._devnull = os.open(os.devnull, os.O_WRONLY)
        except:
            # catch for usage in jupyter notebook, which doesn't support fileno
            self.enable = False
            return

    def __enter__(self):
        if not self.enable:
            return
        try:
            self._newstdout = os.dup(1)
            os.dup2(self._devnull, 1)
            os.close(self._devnull)
            sys.stdout = os.fdopen(self._newstdout, 'w')
        except OSError:
            # "OSError: [Errno 24] Too many open files"
            self.enable = False
            return

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.enable:
            return
        try:
            sys.stdout.close()
            sys.stdout = self._origstdout
            sys.stdout.flush()
            os.dup2(self._oldstdout_fno, 1)
            os.close(self._oldstdout_fno) # Added
        except OSError:
            # "OSError: [Errno 24] Too many open files"
            return

#####################################

MouseEvent = namedtuple('MouseEvent', ['eventType', 'mousePosX', 'mousePosY', 'buttonIndex', 'buttonState'])

def get_mouse_events():
    return list(MouseEvent(*event) for event in p.getMouseEvents(physicsClientId=CLIENT))

def update_viewer():
    # https://docs.python.org/2/library/select.html
    # events = p.getKeyboardEvents() # TODO: only works when the viewer is in focus
    get_mouse_events()
    # for k, v in keys.items():
    #    #p.KEY_IS_DOWN, p.KEY_WAS_RELEASED, p.KEY_WAS_TRIGGERED
    #    if (k == p.B3G_RETURN) and (v & p.KEY_WAS_TRIGGERED):
    #        return
    # time.sleep(1e-3) # Doesn't work
    # disable_gravity()


def wait_for_duration(duration): #, dt=0):
    from pybullet_planning.utils.debug_utils import elapsed_time
    t0 = time.time()
    while elapsed_time(t0) <= duration:
        update_viewer()


def simulate_for_duration(duration):
    dt = get_time_step()
    for i in range(int(duration / dt)):
        step_simulation()


def get_time_step():
    # {'gravityAccelerationX', 'useRealTimeSimulation', 'gravityAccelerationZ', 'numSolverIterations',
    # 'gravityAccelerationY', 'numSubSteps', 'fixedTimeStep'}
    return p.getPhysicsEngineParameters(physicsClientId=CLIENT)['fixedTimeStep']


def enable_separating_axis_test():
    p.setPhysicsEngineParameter(enableSAT=1, physicsClientId=CLIENT)
    #p.setCollisionFilterPair()
    #p.setCollisionFilterGroupMask()
    #p.setInternalSimFlags()
    # enableFileCaching: Set to 0 to disable file caching, such as .obj wavefront file loading
    #p.getAPIVersion() # TODO: check that API is up-to-date
    #p.isNumpyEnabled()


def simulate_for_sim_duration(sim_duration, real_dt=0, frequency=INF):
    t0 = time.time()
    sim_dt = get_time_step()
    sim_time = 0
    last_print = 0
    while sim_time < sim_duration:
        if frequency < (sim_time - last_print):
            print('Sim time: {:.3f} | Real time: {:.3f}'.format(sim_time, elapsed_time(t0)))
            last_print = sim_time
        step_simulation()
        sim_time += sim_dt
        time.sleep(real_dt)


def wait_for_user(message='Press enter to continue'):
    from pybullet_planning.interfaces.env_manager.simulation import has_gui
    if has_gui() and is_darwin():
        # OS X doesn't multi-thread the OpenGL visualizer
        #wait_for_interrupt()
        return threaded_input(message)
    return user_input(message)


def wait_if_gui(*args, **kwargs):
    from pybullet_planning.interfaces.env_manager.simulation import has_gui
    if has_gui():
        wait_for_user(*args, **kwargs)


def is_unlocked():
    return CLIENTS[CLIENT] is True


def wait_if_unlocked(*args, **kwargs):
    if is_unlocked():
        wait_for_user(*args, **kwargs)


def wait_for_interrupt(max_time=np.inf):
    """
    Hold Ctrl to move the camera as well as zoom
    """
    print('Press Ctrl-C to continue')
    try:
        wait_for_duration(max_time)
    except KeyboardInterrupt:
        pass
    finally:
        print()


def step_simulation():
    """stepSimulation will perform all the actions in a single forward dynamics simulation step
    such as collision detection, constraint solving and integration. The default timestep is
    1/240 second, it can be changed using the setTimeStep or setPhysicsEngineParameter API.

    https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/edit#heading=h.czaspku18mzs

    Note: This also forces pybullet to update its bounding volume hierarchy. Ideally one would do this without calling
    the physics simulator. But this is the only workaround that we've found so far.
    """
    p.stepSimulation(physicsClientId=CLIENT)


def threaded_input(*args, **kwargs):
    # OS X doesn't multi-thread the OpenGL visualizer
    # http://openrave.org/docs/0.8.2/_modules/openravepy/misc/#SetViewerUserThread
    # https://github.com/bulletphysics/bullet3/blob/master/examples/pybullet/examples/userData.py
    # https://github.com/bulletphysics/bullet3/tree/master/examples/ExampleBrowser
    #from pybullet_utils import bullet_client
    #from pybullet_utils.bullet_client import BulletClient
    #server = bullet_client.BulletClient(connection_mode=p.SHARED_MEMORY_SERVER) # GUI_SERVER
    #sim_id = p.connect(p.GUI)
    #print(dir(server))
    #client = bullet_client.BulletClient(connection_mode=p.SHARED_MEMORY)
    #sim_id = p.connect(p.SHARED_MEMORY)

    #threading = __import__('threading')
    import threading
    data = []
    thread = threading.Thread(target=lambda: data.append(user_input(*args, **kwargs)), args=[])
    thread.start()
    #threading.enumerate()
    #thread_id = 0
    #for tid, tobj in threading._active.items():
    #    if tobj is thread:
    #        thread_id = tid
    #        break
    try:
        while thread.is_alive():
            update_viewer()
    finally:
        thread.join()
    return data[-1]



from copy import copy
import numpy as np
from collections import namedtuple
import pybullet as p

from pybullet_planning.interfaces.debug_utils.debug_utils import draw_pose, draw_point
from pybullet_planning.interfaces.env_manager.simulation import LockRenderer
from pybullet_planning.interfaces.env_manager.user_io import wait_for_user
from pybullet_planning.utils import BLUE

from pybullet_planning.utils import CLIENT, INFO_FROM_BODY, STATIC_MASS, BASE_LINK, NULL_ID, LOGGER, OBJ_MESH_CACHE
from pybullet_planning.interfaces.env_manager.pose_transformation import Pose, Point, Euler
from pybullet_planning.interfaces.env_manager.pose_transformation import euler_from_quat, base_values_from_pose, \
    quat_from_euler, z_rotation, get_pose, set_pose, apply_affine, unit_pose, multiply, invert
from pybullet_planning.interfaces.env_manager.shape_creation import create_obj, get_collision_data, clone_collision_shape, clone_visual_shape, get_model_info, get_data_pose

from pybullet_planning.interfaces.robots.dynamics import get_mass, get_dynamics_info, get_local_link_pose
from pybullet_planning.interfaces.robots.joint import JOINT_TYPES, get_joint_name, get_joint_type, get_num_joints, is_circular, get_joint_limits, is_fixed, \
    get_joint_info, get_joint_positions, get_joints, is_movable, set_joint_positions, get_movable_joints
from pybullet_planning.interfaces.robots.link import get_links, parent_joint_from_link, get_link_name, get_link_parent, get_link_pose, get_all_links

#####################################
# Bodies

def get_bodies():
    return [p.getBodyUniqueId(i, physicsClientId=CLIENT)
            for i in range(p.getNumBodies(physicsClientId=CLIENT))]

BodyInfo = namedtuple('BodyInfo', ['base_name', 'body_name'])

def get_body_info(body):
    return BodyInfo(*p.getBodyInfo(body, physicsClientId=CLIENT))

def get_base_name(body):
    return get_body_info(body).base_name.decode(encoding='UTF-8')

def get_body_name(body):
    return get_body_info(body).body_name.decode(encoding='UTF-8')

def get_name(body):
    name = get_body_name(body)
    if name == '':
        name = 'body'
        return '{}{}'.format(name, int(body))
    else:
        return name

def has_body(name):
    try:
        body_from_name(name)
    except ValueError:
        return False
    return True

def body_from_name(name):
    for body in get_bodies():
        if get_body_name(body) == name:
            return body
    raise ValueError(name)

def remove_body(body):
    if (CLIENT, body) in INFO_FROM_BODY:
        del INFO_FROM_BODY[CLIENT, body]
    return p.removeBody(body, physicsClientId=CLIENT)

def get_point(body):
    return get_pose(body)[0]

def get_quat(body):
    return get_pose(body)[1] # [x,y,z,w]

def get_euler(body):
    return euler_from_quat(get_quat(body))

def get_base_values(body):
    return base_values_from_pose(get_pose(body))

def set_point(body, point):
    set_pose(body, (point, get_quat(body)))

def set_quat(body, quat):
    set_pose(body, (get_point(body), quat))

def set_euler(body, euler):
    set_quat(body, quat_from_euler(euler))

def pose_from_pose2d(pose2d):
    x, y, theta = pose2d
    return Pose(Point(x=x, y=y), Euler(yaw=theta))

def set_base_values(body, values):
    _, _, z = get_point(body)
    x, y, theta = values
    set_point(body, (x, y, z))
    set_quat(body, z_rotation(theta))

def get_velocity(body):
    linear, angular = p.getBaseVelocity(body, physicsClientId=CLIENT)
    return linear, angular # [x,y,z], [wx,wy,wz]

def set_velocity(body, linear=None, angular=None):
    if linear is not None:
        p.resetBaseVelocity(body, linearVelocity=linear, physicsClientId=CLIENT)
    if angular is not None:
        p.resetBaseVelocity(body, angularVelocity=angular, physicsClientId=CLIENT)

def is_rigid_body(body):
    for joint in get_joints(body):
        if is_movable(body, joint):
            return False
    return True

def is_fixed_base(body):
    return get_mass(body) == STATIC_MASS

def dump_body(body, fixed=False):
    print('Body id: {} | Name: {} | Rigid: {} | Fixed: {}'.format(
        body, get_body_name(body), is_rigid_body(body), is_fixed_base(body)))
    for joint in get_joints(body):
        if fixed or is_movable(body, joint):
            print('Joint id: {} | Name: {} | Type: {} | Circular: {} | Limits: {}'.format(
                joint, get_joint_name(body, joint), JOINT_TYPES[get_joint_type(body, joint)],
                is_circular(body, joint), get_joint_limits(body, joint)))
    link = NULL_ID
    print('Link id: {} | Name: {} | Mass: {} | Collision: {} | Visual: {}'.format(
        link, get_base_name(body), get_mass(body),
        len(get_collision_data(body, link)), NULL_ID)) # len(get_visual_data(body, link))))
    for link in get_links(body):
        joint = parent_joint_from_link(link)
        joint_name = JOINT_TYPES[get_joint_type(body, joint)] if is_fixed(body, joint) else get_joint_name(body, joint)
        print('Link id: {} | Name: {} | Joint: {} | Parent: {} | Mass: {} | Collision: {} | Visual: {}'.format(
            link, get_link_name(body, link), joint_name,
            get_link_name(body, get_link_parent(body, link)), get_mass(body, link),
            len(get_collision_data(body, link)), NULL_ID)) # len(get_visual_data(body, link))))
        #print(get_joint_parent_frame(body, link))
        #print(map(get_data_geometry, get_visual_data(body, link)))
        #print(map(get_data_geometry, get_collision_data(body, link)))

def dump_world():
    """print out information for all bodies that're currently in the scene

    """
    for body in get_bodies():
        dump_body(body)
        print()

def clone_body(body, links=None, collision=True, visual=True, client=None):
    # ! links must be in a consecutive list of integers, so no input like `[link_2, link_4]`
    from pybullet_planning.utils import get_client
    # TODO: names are not retained
    # TODO: error with createMultiBody link poses on PR2
    # localVisualFrame_position: position of local visual frame, relative to link/joint frame
    # localVisualFrame orientation: orientation of local visual frame relative to link/joint frame
    # parentFramePos: joint position in parent frame
    # parentFrameOrn: joint orientation in parent frame
    # ! the following get around a bug when cloning from an object created from obj file with a non-one scale
    model_info = get_model_info(body)
    if model_info is not None and model_info.path.endswith('.obj'):
        new_body = create_obj(model_info.path, scale=model_info.scale, collision=collision)
        set_pose(new_body, get_pose(body))
        INFO_FROM_BODY[CLIENT, new_body] = copy(model_info)
        return new_body

    client = get_client(client) # client is the new client for the body
    if links is None:
        links = get_links(body)
    if BASE_LINK in links:
        # ! links should not include BASE_LINK, otherwise get_joint_info will fail
        links.pop(links.index(BASE_LINK))
    #movable_joints = [joint for joint in links if is_movable(body, joint)]
    new_from_original = {}
    base_link = get_link_parent(body, links[0]) if links else BASE_LINK
    new_from_original[base_link] = NULL_ID

    masses = []
    collision_shapes = []
    visual_shapes = []
    positions = [] # list of local link positions, with respect to parent
    orientations = [] # list of local link orientations, w.r.t. parent
    inertial_positions = [] # list of local inertial frame pos. in link frame
    inertial_orientations = [] # list of local inertial frame orn. in link frame
    parent_indices = []
    joint_types = []
    joint_axes = []
    for i, link in enumerate(links):
        new_from_original[link] = i
        joint_info = get_joint_info(body, link)
        dynamics_info = get_dynamics_info(body, link)
        masses.append(dynamics_info.mass)
        collision_shapes.append(clone_collision_shape(body, link, client) if collision else NULL_ID)
        visual_shapes.append(clone_visual_shape(body, link, client) if visual else NULL_ID)
        point, quat = get_local_link_pose(body, link)
        positions.append(point)
        orientations.append(quat)
        inertial_positions.append(dynamics_info.local_inertial_pos)
        inertial_orientations.append(dynamics_info.local_inertial_orn)
        parent_indices.append(new_from_original[joint_info.parentIndex] + 1) # TODO: need the increment to work
        joint_types.append(joint_info.jointType)
        joint_axes.append(joint_info.jointAxis)
    # https://github.com/bulletphysics/bullet3/blob/9c9ac6cba8118544808889664326fd6f06d9eeba/examples/pybullet/gym/pybullet_utils/urdfEditor.py#L169

    base_dynamics_info = get_dynamics_info(body, base_link)
    base_point, base_quat = get_link_pose(body, base_link)
    new_body = p.createMultiBody(baseMass=base_dynamics_info.mass,
                                 baseCollisionShapeIndex=clone_collision_shape(body, base_link, client) if collision else NULL_ID,
                                 baseVisualShapeIndex=clone_visual_shape(body, base_link, client) if visual else NULL_ID,
                                 basePosition=base_point,
                                 baseOrientation=base_quat,
                                 baseInertialFramePosition=base_dynamics_info.local_inertial_pos,
                                 baseInertialFrameOrientation=base_dynamics_info.local_inertial_orn,
                                 linkMasses=masses,
                                 linkCollisionShapeIndices=collision_shapes,
                                 linkVisualShapeIndices=visual_shapes,
                                 linkPositions=positions,
                                 linkOrientations=orientations,
                                 linkInertialFramePositions=inertial_positions,
                                 linkInertialFrameOrientations=inertial_orientations,
                                 linkParentIndices=parent_indices,
                                 linkJointTypes=joint_types,
                                 linkJointAxis=joint_axes,
                                 physicsClientId=client)
    #set_configuration(new_body, get_joint_positions(body, movable_joints)) # Need to use correct client
    for joint, value in zip(range(len(links)), get_joint_positions(body, links)):
        # TODO: check if movable?
        p.resetJointState(new_body, joint, value, targetVelocity=0, physicsClientId=client)
    # * cache when cloning from objects created from .obj mesh file
    model_info = get_model_info(body)
    if model_info is not None:
        INFO_FROM_BODY[CLIENT, new_body] = copy(model_info)
    return new_body

def clone_world(client=None, exclude=[]):
    from pybullet_planning.interfaces.env_manager.simulation import has_gui
    visual = has_gui(client)
    mapping = {}
    for body in get_bodies():
        if body not in exclude:
            new_body = clone_body(body, collision=True, visual=visual, client=client)
            mapping[body] = new_body
    return mapping

def set_color(body, color, link=None, shape_index=NULL_ID):
    """
    Experimental for internal use, recommended ignore shapeIndex or leave it NULL_ID.
    Intention was to let you pick a specific shape index to modify,
    since URDF (and SDF etc) can have more than 1 visual shape per link.
    This shapeIndex matches the list ordering returned by getVisualShapeData.
    :param body:
    :param link:
    :param shape_index:
    :return:
    """
    # specularColor
    links = get_all_links(body) if link is None else [link]
    return_flags = []
    for body_link in links:
        return_flags.append(p.changeVisualShape(body, body_link, shapeIndex=shape_index, rgbaColor=color,
                               #textureUniqueId=None, specularColor=None,
                               physicsClientId=CLIENT))
    return all(return_flags)

def set_texture(body, texture=None, link=BASE_LINK, shape_index=NULL_ID):
    if texture is None:
        texture = NULL_ID
    return p.changeVisualShape(body, link, shapeIndex=shape_index, textureUniqueId=texture,
                               physicsClientId=CLIENT)

def get_body_collision_vertices(body):
    """get body link collision body's vertices in its current configuration

    Parameters
    ----------
    body : int

    Returns
    -------
    dict
        {link_id : list[point 3d position]}
    """
    from .link import get_com_pose

    body_vertices_from_link = {}
    # joints = get_movable_joints(body)
    # conf = get_joint_positions(body, joints)
    # set_joint_positions(body, get_movable_joints(body), conf)
    body_links = get_links(body)
    # * get links except for BASE_LINK
    for body_link in body_links:
        local_from_vertices = vertices_from_rigid(body, body_link, collision=True)
        world_from_com = get_com_pose(body, body_link)
        body_vertices_from_link[body_link] = apply_affine(world_from_com, local_from_vertices)
    # * base link handling
    base_local_from_vertices = vertices_from_rigid(body, BASE_LINK, collision=True)
    world_from_current_pose = get_link_pose(body, BASE_LINK)
    body_vertices_from_link[BASE_LINK] = apply_affine(world_from_current_pose, base_local_from_vertices)

    return body_vertices_from_link

def vertices_from_link(body, link, collision=True):
    """get body link geometry's vertices in local frame

    Parameters
    ----------
    body : int
    link : int

    Returns
    -------
    list(np.array())
        vertices of the geometry
    """
    from pybullet_planning.interfaces.env_manager.shape_creation import vertices_from_data, get_visual_data
    # In local frame
    vertices = []
    # ! PyBullet creates multiple collision elements (with unknown_file) when nonconvex
    get_data = get_collision_data if collision else get_visual_data
    for data in get_data(body, link):
        vertices.extend(vertices_from_data(data, body))
    return vertices

def vertices_from_rigid(body, link=BASE_LINK, collision=True):
    """get all vertices of given body (collision body) in its local frame.
    This is a more stable version compared to `vertices_from_link` when getting vertices from a rigid body
    without joints.

    PyBullet creates multiple collision elements (with unknown_file) when nonconvex and those objects does not
    have `unknown` filename. This function has a catch for such cases by loading vertices directly from obj file.

    Parameters
    ----------
    body : int
        [description]
    link : [type], optional
        if BASE_LINK, we assume the body is single-linked, by default BASE_LINK

    Returns
    -------
    list of three-float lists
        body vertices
    """
    import os
    from pybullet_planning.interfaces.geometry.mesh import read_obj
    from pybullet_planning.interfaces.env_manager import get_model_info
    # from pybullet_planning.interfaces.robots.link import get_num_links
    # assert implies(link == BASE_LINK, get_num_links(body) == 0), 'body {} has links {}'.format(body, get_all_links(body))
    # try:
    return vertices_from_link(body, link, collision=collision)
    # except RuntimeError as e:
    #     # ! this
    #     info = get_model_info(body)
    #     assert info is not None
    #     _, ext = os.path.splitext(info.path)
    #     if ext == '.obj':
    #         if info.path not in OBJ_MESH_CACHE:
    #             OBJ_MESH_CACHE[info.path] = read_obj(info.path, decompose=False)
    #         mesh = OBJ_MESH_CACHE[info.path]
    #         vertices = [[v[i]*info.scale for i in range(3)] for v in mesh.vertices]
    #         # LOGGER.debug('cloned from obj in vertices_from_rigid')
    #     else:
    #         raise e
    # return vertices

def approximate_as_prism(body, body_pose=unit_pose(), link=BASE_LINK, **kwargs):
    """get the AABB bounding box of a body

    Note: the generated AABB is not truly axis-aligned, bounding box under world axis x, y

    Parameters
    ----------
    body : int
        pb body's index
    body_pose : [type], optional
        [description], by default unit_pose()

    Returns
    -------
    tuple of (Point, float)
        bounding box center and extent length
    """
    from pybullet_planning.interfaces.geometry.bounding_box import aabb_from_points, get_aabb_center, get_aabb_extent
    # TODO: make it just orientation
    vertices = apply_affine(body_pose, vertices_from_rigid(body, link=link, **kwargs))
    aabb = aabb_from_points(vertices)
    return get_aabb_center(aabb), get_aabb_extent(aabb)
    #with PoseSaver(body):
    #    set_pose(body, body_pose)
    #    set_velocity(body, linear=np.zeros(3), angular=np.zeros(3))
    #    return get_center_extent(body, **kwargs)

def approximate_as_cylinder(body, **kwargs):
    """get the bounding cylinder of the AABB bounding box of a body

    Note: the generated AABB is not truly axis-aligned, bounding box under world axis x, y.
    Thus, the estimated diameter and height might not be accurate in the longitude axis aligning sense.

    Parameters
    ----------
    body : int
        [description]

    Returns
    -------
    Point, tuple of (diameter, height)
        [description]
    """
    center, (width, length, height) = approximate_as_prism(body, **kwargs)
    diameter = np.sqrt(width**2 + length**2) # TODO: check that these are close
    diameter = (width + length) / 2 # TODO: check that these are close
    return center, (diameter, height)

#####################################

def load_model(rel_path, pose=None, **kwargs):
    from pybullet_planning.interfaces.env_manager.simulation import get_model_path, add_data_path, load_pybullet
    # TODO: error with loadURDF when loading MESH visual and CYLINDER collision
    abs_path = get_model_path(rel_path)
    add_data_path()
    #with LockRenderer():
    body = load_pybullet(abs_path, **kwargs)
    if pose is not None:
        set_pose(body, pose)
    return body



from collections import namedtuple
from collections.abc import Mapping
from heapq import heappush, heappop


class Vertex(object):

    def __init__(self, value):
        self.value = value
        self.edges = []

    def __repr__(self):
        return self.__class__.__name__ + '(' + str(self.value) + ')'


class Edge(object):

    def __init__(self, v1, v2, value, cost):
        self.v1, self.v2 = v1, v2
        self.v1.edges.append(self)
        self.value = value
        self.cost = cost

    def __repr__(self):
        return self.__class__.__name__ + '(' + str(self.v1.value) + ' -> ' + str(self.v2.value) + ')'

SearchNode = namedtuple('SearchNode', ['cost', 'edge'])


class Graph(Mapping, object):

    def __init__(self):
        self.vertices = {}
        self.edges = []

    def __getitem__(self, value):
        return self.vertices[value]

    def __len__(self):
        return len(self.vertices)

    def __iter__(self):
        return iter(self.vertices)

    def __call__(self, value1, value2):  # TODO - goal_test
        if value1 not in self or value2 not in self:
            return None

        start, goal = self[value1], self[value2]
        queue = [(0, start)]
        nodes, processed = {start: SearchNode(0, None)}, set()

        def retrace(v):
            edge = nodes[v].edge
            if edge is None:
                return [v.value], []
            vertices, edges = retrace(edge.v1)
            return vertices + [v.value], edges + [edge.value]

        while len(queue) != 0:
            _, cv = heappop(queue)
            if cv in processed:
                continue
            processed.add(cv)
            if cv == goal:
                return retrace(cv)
            for edge in cv.edges:
                cost = nodes[cv].cost + edge.cost
                if edge.v2 not in nodes or cost < nodes[edge.v2].cost:
                    nodes[edge.v2] = SearchNode(cost, edge)
                    heappush(queue, (cost, edge.v2))
        return None

    def add(self, value):
        if value not in self:
            self.vertices[value] = Vertex(value)
        return self.vertices[value]

    def connect(self, value1, value2, edge_value=None, edge_cost=1):
        v1, v2 = self.add(value1), self.add(value2)
        edge = Edge(v1, v2, edge_value, edge_cost)
        self.edges.append(edge)
        return edge



"""
********************************************************************************
pybullet_planning
********************************************************************************

.. currentmodule:: pybullet_planning

This library is a suite of utility functions to facilitate robotic planning related research on the
`pybullet <https://github.com/bulletphysics/bullet3>`_ physics simulation engine.

.. toctree::
    :maxdepth: 1

    pybullet_planning.interfaces
    pybullet_planning.primitives
    pybullet_planning.motion_planners
    pybullet_planning.utils

"""

from .__version__ import __author__, __author_email__, __copyright__, __description__, __license__, \
    __title__, __url__, __version__

from .utils import *
from .interfaces import *
from .motion_planners import *
from .primitives import *

__all__ = ['__author__', '__author_email__', '__copyright__', '__description__', '__license__', '__title__', '__url__', '__version__']



import os
import warnings
from itertools import tee
from copy import copy
from collections import namedtuple
import numpy as np
import pybullet as p

from pybullet_planning.interfaces.env_manager.pose_transformation import get_distance
from pybullet_planning.utils import MAX_DISTANCE, EPS, INF
from pybullet_planning.interfaces.robots.joint import get_joint_positions, get_custom_limits, get_movable_joints, set_joint_positions, \
    get_configuration, get_custom_max_velocity
from pybullet_planning.interfaces.robots.collision import get_collision_fn
from pybullet_planning.interfaces.robots.body import clone_body, remove_body, get_link_pose
from .ladder_graph import LadderGraph, EdgeBuilder, append_ladder_graph, DEFAULT_DTHETA
from .dag_search import DAGSearch

#####################################

NullSpace = namedtuple('Nullspace', ['lower', 'upper', 'range', 'rest'])

def get_null_space(robot, joints, custom_limits={}):
    rest_positions = get_joint_positions(robot, joints)
    lower, upper = get_custom_limits(robot, joints, custom_limits)
    lower = np.maximum(lower, -10*np.ones(len(joints)))
    upper = np.minimum(upper, +10*np.ones(len(joints)))
    joint_ranges = 10*np.ones(len(joints))
    return NullSpace(list(lower), list(upper), list(joint_ranges), list(rest_positions))

def plan_cartesian_motion(robot, first_joint, target_link, waypoint_poses,
                          max_iterations=200, custom_limits={}, get_sub_conf=False, **kwargs):
    """Compute a joint trajectory for a given sequence of workspace poses. Only joint limit is considered.
    Collision checking using `get_collision_fn` is often performed on the path computed by this function.

    Parameters
    ----------
    robot : int
        robot body index
    first_joint : int
        the first joint index in the kinematics chain.
    target_link : int
        end effector link index.
    waypoint_poses : a list of Pose
        a list of end effector workspace poses in the world coord.
    max_iterations : int, optional
        [description], by default 200
    custom_limits : dict, optional
        [description], by default {}
    get_sub_conf : bool, optional
        return sub-kinematics chain configuration if set to True, by default False

    Returns
    -------
    [type]
        [description]

    Example
    -------
    ```
    ik_joints = joints_from_names(robot_uid, ik_joint_names)
    ik_tool_link = link_from_name(robot_uid, tool_link_name)
    cart_conf_vals = plan_cartesian_motion(robot_uid, ik_joints[0], ik_tool_link, world_from_ee_poses)
    ```
    """
    from pybullet_planning.interfaces.env_manager.pose_transformation import all_between
    from pybullet_planning.interfaces.robots.link import get_link_subtree, prune_fixed_joints
    from pybullet_planning.interfaces.kinematics import inverse_kinematics_helper, is_pose_close

    # TODO: fix stationary joints
    # TODO: pass in set of movable joints and take least common ancestor
    # TODO: update with most recent bullet updates
    # https://github.com/bulletphysics/bullet3/blob/master/examples/pybullet/examples/inverse_kinematics.py
    # https://github.com/bulletphysics/bullet3/blob/master/examples/pybullet/examples/inverse_kinematics_husky_kuka.py
    # TODO: plan a path without needing to following intermediate waypoints

    lower_limits, upper_limits = get_custom_limits(robot, get_movable_joints(robot), custom_limits)
    # TODO might have problem in a dual-arm setting, where the two arms share the same first joint
    selected_links = get_link_subtree(robot, first_joint) # TODO: child_link_from_joint?
    selected_movable_joints = prune_fixed_joints(robot, selected_links)
    assert(target_link in selected_links)
    selected_target_link = selected_links.index(target_link)
    sub_robot = clone_body(robot, links=selected_links, visual=False, collision=False) # TODO: joint limits
    sub_movable_joints = get_movable_joints(sub_robot)
    #null_space = get_null_space(robot, selected_movable_joints, custom_limits=custom_limits)
    null_space = None

    solutions = []
    for target_pose in waypoint_poses:
        for iteration in range(max_iterations):
            sub_kinematic_conf = inverse_kinematics_helper(sub_robot, selected_target_link, target_pose, null_space=null_space)
            if sub_kinematic_conf is None:
                remove_body(sub_robot)
                return None
            set_joint_positions(sub_robot, sub_movable_joints, sub_kinematic_conf)
            if is_pose_close(get_link_pose(sub_robot, selected_target_link), target_pose, **kwargs):
                set_joint_positions(robot, selected_movable_joints, sub_kinematic_conf)
                kinematic_conf = get_configuration(robot)
                if not all_between(lower_limits, kinematic_conf, upper_limits):
                    #movable_joints = get_movable_joints(robot)
                    #print([(get_joint_name(robot, j), l, v, u) for j, l, v, u in
                    #       zip(movable_joints, lower_limits, kinematic_conf, upper_limits) if not (l <= v <= u)])
                    #print("Limits violated")
                    #wait_for_user()
                    remove_body(sub_robot)
                    return None
                #print("IK iterations:", iteration)
                if not get_sub_conf:
                    solutions.append(kinematic_conf)
                else:
                    solutions.append(sub_kinematic_conf)
                break
        else:
            remove_body(sub_robot)
            return None
    remove_body(sub_robot)
    return solutions

def sub_inverse_kinematics(robot, first_joint, target_link, target_pose, **kwargs):
    solutions = plan_cartesian_motion(robot, first_joint, target_link, [target_pose], **kwargs)
    if solutions:
        return solutions[0]
    return None

#####################################

MAX_SAMPLE_ITER = int(1e4)

def plan_cartesian_motion_lg(robot, joints, waypoint_poses, sample_ik_fn=None, collision_fn=None, sample_ee_fn=None,
    max_sample_ee_iter=MAX_SAMPLE_ITER, jump_threshold=None, enforce_start_conf=True, **kwargs):
    """ladder graph cartesian planning, better leveraging ikfast for sample_ik_fn

    Parameters
    ----------
    robot : [type]
        [description]
    joints : [type]
        [description]
    waypoint_poses : [type]
        [description]
    sample_ik_fn : [type], optional
        [description], by default None
    collision_fn : [type], optional
        [description], by default None
    ee_sample_fn : [type], optional
        please please please remember to put an end to the sampling loop!
    jump_threshold : [type], optional

    Returns
    -------
    [type]
        [description]
    """
    from pybullet_planning.interfaces.env_manager.user_io import wait_for_user
    from pybullet_planning.interfaces.env_manager.savers import WorldSaver

    assert sample_ik_fn is not None, 'Sample fn must be specified!'

    # TODO sanity check samplers
    with WorldSaver():
        ik_sols = [[] for _ in range(len(waypoint_poses))]
        for i, task_pose in enumerate(waypoint_poses):
            candidate_poses = [task_pose]
            if sample_ee_fn is not None:
                # extra dof release, copy to reuse generator
                current_ee_fn = copy(sample_ee_fn)
                cnt = 0
                for p in current_ee_fn(task_pose):
                    if cnt > max_sample_ee_iter:
                        warnings.warn('EE dof release generator is called over {} times, likely that you forget to put an exit in the generator. ' + \
                            'We stop generating here for you.')
                        break
                    candidate_poses.append(p)
                    cnt += 1
            for ee_pose in candidate_poses:
                conf_list = sample_ik_fn(ee_pose)
                if collision_fn is not None:
                    conf_list = [conf for conf in conf_list if conf and not collision_fn(conf, **kwargs)]
                ik_sols[i].extend(conf_list)

    # assemble the ladder graph
    dof = len(joints)
    graph = LadderGraph(dof)
    graph.resize(len(ik_sols))

    # assign rung data
    for pt_id, ik_confs_pt in enumerate(ik_sols):
        graph.assign_rung(pt_id, ik_confs_pt)

    joint_jump_threshold = None
    if jump_threshold is not None:
        joint_jump_threshold = []
        for joint in joints:
            if joint in jump_threshold:
                joint_jump_threshold.append(jump_threshold[joint])
            else:
                joint_jump_threshold.append(DEFAULT_DTHETA)

    # build edges within current pose family
    for i in range(graph.get_rungs_size()-1):
        st_rung_id = i
        end_rung_id = i + 1
        jt1_list = graph.get_data(st_rung_id)
        jt2_list = graph.get_data(end_rung_id)
        st_size = graph.get_rung_vert_size(st_rung_id)
        end_size = graph.get_rung_vert_size(end_rung_id)
        # if st_size == 0 or end_size == 0:
        #     print(ik_sols)

        if st_size == 0:
            print('Ladder graph not valid: rung {}/{} is a zero size rung'.format(st_rung_id, graph.get_rungs_size()))
            return None, None
        if end_size == 0:
            print('Ladder graph not valid: rung {}/{} is a zero size rung'.format(end_rung_id, graph.get_rungs_size()))
            return None, None

        # TODO: preference_cost using pose deviation
        # fully-connected ladder graph
        edge_builder = EdgeBuilder(st_size, end_size, dof,
            jump_threshold=joint_jump_threshold, **kwargs)
        for k in range(st_size):
            st_jt_id = k * dof
            for j in range(end_size):
                end_jt_id = j * dof
                edge_builder.consider(jt1_list[st_jt_id : st_jt_id+dof], jt2_list[end_jt_id : end_jt_id+dof], j)
            edge_builder.next(k)

        # print('({}) rung'.format(i))
        # print('max_dtheta_: ', edge_builder.max_dtheta_)
        # print('edge_scratch_: ', edge_builder.edge_scratch_)
        # print('result: ', edge_builder.result)
        # TODO: more report information here
        if not edge_builder.has_edges:
            print('Ladder graph: no edge built between {}-{} | joint threshold: {}, max delta jt: {}'.format(
                st_rung_id, end_rung_id, joint_jump_threshold, edge_builder.max_dtheta_))
            return None, None

        edges = edge_builder.result
        graph.assign_edges(i, edges)

    # * use current conf in the env as start_conf
    if enforce_start_conf:
        start_conf = get_joint_positions(robot, joints)
        st_graph = LadderGraph(graph.dof)
        st_graph.resize(1)
        st_graph.assign_rung(0, [start_conf])
        unified_graph = append_ladder_graph(st_graph, graph, jump_threshold=joint_jump_threshold)
        if unified_graph is None:
            return None, None
    else:
        unified_graph = graph

    # perform DAG search
    dag_search = DAGSearch(unified_graph)
    min_cost = dag_search.run()
    # list of confs
    path = dag_search.shortest_path()
    if enforce_start_conf:
        # delete the start conf
        del path[0]

    assert len(path) == len(waypoint_poses)
    if len(path) == 0 or min_cost == np.inf:
        return None, None
    return path, min_cost



import numpy as np
import pybullet as p

# constants
INF = np.inf
PI = np.pi
EPS = 1e-12
CIRCULAR_LIMITS = -PI, PI
UNBOUNDED_LIMITS = -INF, INF
DEFAULT_TIME_STEP = 1./240. # seconds

# pybullet setup parameters
CLIENTS = {} # TODO: rename to include locked
CLIENT = 0

INFO_FROM_BODY = {}

# colors
RED = (1, 0, 0, 1)
GREEN = (0, 1, 0, 1)
BLUE = (0, 0, 1, 1)
BLACK = (0, 0, 0, 1)
WHITE = (1, 1, 1, 1)
BROWN = (0.396, 0.263, 0.129, 1)
TAN = (0.824, 0.706, 0.549, 1)
GREY = (0.5, 0.5, 0.5, 1)
YELLOW = (1, 1, 0, 1)

GRAVITY = 9.8

# links
BASE_LINK = -1
STATIC_MASS = 0

# geometry
DEFAULT_MESH = ''
DEFAULT_EXTENTS = [1, 1, 1]
DEFAULT_RADIUS = 0.5
DEFAULT_HEIGHT = 1
DEFAULT_SCALE = [1, 1, 1]
DEFAULT_NORMAL = [0, 0, 1]

# planning
DEFAULT_RESOLUTION = 0.05

# Collision
# MAX_DISTANCE = 1e-3
# Used in collision checking query, e.g. pybullet.getClosestPoint
# If the distance between objects exceeds this maximum distance, no points may be returned.
MAX_DISTANCE = 0.

UNKNOWN_FILE = 'unknown_file'
NULL_ID = -1
TEMP_DIR = 'temp/'

OBJ_MESH_CACHE = {}
"""a cache keeping track of loaded obj objects

"""

CARTESIAN_TYPES = {
    'x': (p.JOINT_PRISMATIC, [1, 0, 0]),
    'y': (p.JOINT_PRISMATIC, [0, 1, 0]),
    'z': (p.JOINT_PRISMATIC, [0, 0, 1]),
    'roll': (p.JOINT_REVOLUTE, [1, 0, 0]),
    'pitch': (p.JOINT_REVOLUTE, [0, 1, 0]),
    'yaw': (p.JOINT_REVOLUTE, [0, 0, 1]),
}

T2 = ['x', 'y']
T3 = ['x', 'y', 'z']

SE2 = T2 + ['yaw']
SE3 = T3 + ['roll', 'pitch', 'yaw']

def get_client(client=None):
    if client is None:
        return CLIENT
    return client

def set_client(client):
    global CLIENT
    CLIENT = client

DATE_FORMAT = '%y-%m-%d_%H-%M-%S'



import numpy as np

from pybullet_planning.utils import CIRCULAR_LIMITS

#####################################
# Reachability

def sample_reachable_base(robot, point, reachable_range=(0.25, 1.0)):
    from pybullet_planning.interfaces.env_manager.pose_transformation import unit_from_theta, point_from_pose
    radius = np.random.uniform(*reachable_range)
    x, y = radius*unit_from_theta(np.random.uniform(-np.pi, np.pi)) + point[:2]
    yaw = np.random.uniform(*CIRCULAR_LIMITS)
    base_values = (x, y, yaw)
    #set_base_values(robot, base_values)
    return base_values

def uniform_pose_generator(robot, gripper_pose, **kwargs):
    from pybullet_planning.interfaces.env_manager.pose_transformation import point_from_pose
    point = point_from_pose(gripper_pose)
    while True:
        base_values = sample_reachable_base(robot, point, **kwargs)
        if base_values is None:
            break
        yield base_values
        #set_base_values(robot, base_values)
        #yield get_pose(robot)



from itertools import combinations, product, permutations
from scipy.spatial.kdtree import KDTree

import numpy as np

from .utils import INF, compute_path_cost
from pybullet_planning.interfaces.env_manager.pose_transformation import get_distance

# TODO: shared roadmap for multi-query

def compute_median_distance(path1, path2):
    differences = [get_distance(q1, q2) for q1, q2 in product(path1, path2)]
    return np.median(differences)


def compute_minimax_distance(path1, path2):
    overall_distance = 0.
    for path, other in permutations([path1, path2]):
        tree = KDTree(other)
        for q1 in path:
            #closest_distance = min(get_distance(q1, q2) for q2 in other)
            closest_distance, closest_index = tree.query(q1, k=1, eps=0.)
            overall_distance = max(overall_distance, closest_distance)
    return overall_distance


def compute_portfolio_distance(path1, path2, min_distance=0.):
    # TODO: generic distance_fn
    # TODO: min_distance from stats about the portfolio
    distance = compute_minimax_distance(path1, path2)
    if distance < min_distance:
        return 0.
    return sum(compute_path_cost(path, get_distance) for path in [path1, path2])


def score_portfolio(portfolio, **kwargs):
    # TODO: score based on collision_fn voxel overlap at different resolutions
    score_fn = compute_minimax_distance # compute_median_distance | compute_minimax_distance | compute_portfolio_distance
    score = INF
    for path1, path2 in combinations(portfolio, r=2):
        score = min(score, score_fn(path1, path2, **kwargs))
    return score


def exhaustively_select_portfolio(candidates, k=10, **kwargs):
    if len(candidates) <= k:
        return candidates
    # TODO: minimum length portfolio such that at nothing is within a certain distance_fn
    best_portfolios, best_score = [], 0
    for portfolio in combinations(candidates, r=k):
        score = score_portfolio(portfolio, **kwargs)
        if score > best_score:
            best_portfolios, best_score = portfolio, score
    return best_portfolios


def greedily_select_portfolio(candidates, k=10):
    # Higher score is better
    if len(candidates) <= k:
        return candidates
    raise NotImplementedError()
    #return best_portfolios




from pybullet_planning.interfaces.env_manager.pose_transformation import multiply, invert, quat_from_matrix, matrix_from_quat, \
    point_from_pose, quat_from_pose, get_distance, unit_pose
from pybullet_planning.interfaces.robots.joint import joints_from_names, get_joint_positions, violates_limits
from pybullet_planning.interfaces.robots.link import get_link_pose, link_from_name

def get_ik_tool_link_pose(fk_fn, robot, ik_joints, base_link, \
                          joint_values=None, use_current=False):
    """Use the given forward_kinematics function to compute ik_tool_link pose
    based on current joint configurations in pybullet.

    Note: The resulting FK pose is relative to the world frame, not the robot's base frame.

    Parameters
    ----------
    fk_fn : function handle
        fk(a 6-list) : point, rot_matrix
    robot : pybullet robot
    ik_joint : list of int
        a list of pybullet joint index that is registered for IK/FK
        can be obtained by e.g.: ``ik_joints = joints_from_names(robot, ik_joint_names)``
    base_link : int
        robot base link index, can be obtained from ``link_from_name(robot, base_link_name)``
    joint_values : list of float
        robot joint values for FK computation, value correponds to ik_joint_names, default to None
    use_current: bool
        if true, use the current configuration in pybullet env for FK, default
        to False

    Returns
    -------
    pybullet Pose
        Pose = (point, quat) = ([x,y,z], [4-list])
    """
    if use_current:
        conf = get_joint_positions(robot, ik_joints)
    else:
        assert joint_values
        conf = joint_values

    base_from_tool = compute_forward_kinematics(fk_fn, conf)
    world_from_base = get_link_pose(robot, base_link)
    return multiply(world_from_base, base_from_tool)


def get_ik_generator(ik_fn, robot, base_link, world_from_tcp, ik_tool_link_from_tcp=None, **kwargs):
    """get an ik generator

    Parameters
    ----------
    ik_fn : function handle
        get_ik(point, rot) : list of jt solutions
        point = [x,y,z]
        rot = 3x3 rotational matrix as a row-major list
    robot : pybullet robot
    base_link : int
        robot base link index, can be obtained from ``link_from_name(robot, base_link_name)``
    world_from_tcp : pybullet pose
        tcp pose in the world frame
    ik_tool_link_from_tcp : pybullet pose
        tcp pose in ik tool link's frame, optional, defaults = None

    Returns
    -------
    generator
        use next() to get solutions
    """
    world_from_base = get_link_pose(robot, base_link)
    base_from_tcp = multiply(invert(world_from_base), world_from_tcp)
    if ik_tool_link_from_tcp:
        base_from_ik_tool_link = multiply(base_from_tcp, invert(ik_tool_link_from_tcp))
    else:
        base_from_ik_tool_link = base_from_tcp
    yield compute_inverse_kinematics(ik_fn, base_from_ik_tool_link, **kwargs)


def sample_tool_ik(ik_fn, robot, ik_joints, world_from_tcp, base_link,
                   ik_tool_link_from_tcp=None,  closest_only=False, get_all=False, **kwargs):
    """ sample ik joints for a given tcp pose in the world frame

    Parameters
    ----------
    ik_fn : function handle
        get_ik(point, rot) : list of jt solutions
        point = [x,y,z]
        rot = 3x3 rotational matrix as a row-major list
    robot : pybullet robot
    ik_joint : list of int
        a list of pybullet joint index that is registered for IK/FK
        can be obtained by e.g.: ``ik_joints = joints_from_names(robot, ik_joint_names)``
    base_link : int
        robot base link index, can be obtained from ``link_from_name(robot, base_link_name)``
    world_from_tcp : pybullet pose
        tcp pose in the world frame
    ik_tool_link_from_tcp : pybullet pose
        tcp pose in ik tool link's frame, optional, defaults to None
    get_all : bool
        get all ik sol computed
    **kwargs : optional arguments
        for solution selection, nearby_conf=True to select the one that's closest
        to current conf in pybullet env

    Returns
    -------
    a list of 6-list
        computed IK solutions that satisfy joint limits
    """
    generator = get_ik_generator(ik_fn, robot, base_link, world_from_tcp, ik_tool_link_from_tcp, **kwargs)
    sols = next(generator)
    if closest_only and sols:
        current_conf = get_joint_positions(robot, ik_joints)
        sols = [min(sols, key=lambda conf: get_distance(current_conf, conf))]
    sols = list(filter(lambda conf: not violates_limits(robot, ik_joints, conf), sols))
    return sols if get_all else select_solution(robot, ik_joints, sols, **kwargs)


def compute_forward_kinematics(fk_fn, conf):
    """use the given fk function to compute FK

    Parameters
    ----------
    fk_fn : function handle
    conf : list
        joint values

    Returns
    -------
    pos: 3-list
        [x,y,z] of the FK solution pose
    quat: 4-list
        quaternion of the FK solution pose
    """
    pose = fk_fn(list(conf))
    pos, rot = pose
    quat = quat_from_matrix(rot) # [X,Y,Z,W]
    return pos, quat


def compute_inverse_kinematics(ik_fn, pose, sampled=[]):
    """compute ik solutions using the given ik function handle

    Parameters
    ----------
    ik_fn : function handle
        get_ik(point, rot) : list of jt solutions
        point = [x,y,z]
        rot = 3x3 rotational matrix as a row-major list
    pose : pybullet pose
        pose of the ik tool link
    sampled : list
        a list of externally sampled solutions that wants to be appended to the
        computed ik solutions

    Returns
    -------
    a list of 6-lists
        a list of ik solutions
    """
    pos = point_from_pose(pose)
    rot = matrix_from_quat(quat_from_pose(pose)).tolist()
    if sampled:
        solutions = ik_fn(list(pos), list(rot), sampled)
    else:
        solutions = ik_fn(list(pos), list(rot))
    if solutions is None:
        return []
    return solutions


def select_solution(body, joints, solutions, nearby_conf=True, random=False, **kwargs):
    """select one joint configuration given a list of them

    Parameters
    ----------
    body : pybullet body
        This can be any pybullet body, including the robot
    joints : a list of pybullet joint
    solutions : a list of float-lists
        joint values
    nearby_conf : bool
        return the joint conf that is closest to the current conf in pybullet env,
        defaults to True
    random : bool
        randomly chose one
    **kwargs : optional
        additional arguments for the cost function when ranking

    Returns
    -------
    a list of lists or one list
        the joint configuration(s)
    """
    if not solutions:
        return None
    if random and not nearby_conf:
        return random.choice(solutions)
    if nearby_conf:
        nearby_conf = get_joint_positions(body, joints)
        # TODO: sort by distance before collision checking
        # TODO: search over neighborhood of sampled joints when nearby_conf != None
        return min(solutions, key=lambda conf: get_distance(nearby_conf, conf, **kwargs))
    else:
        return random.choice(solutions)



import warnings

import numpy as np
from .ladder_graph import LadderGraph, EdgeBuilder

class SolutionRung(object):
    def __init__(self):
        self.distance = []
        self.predecessor = []

    def extract_min(self):
        # min_dist = min(self.distance)
        # min_id = self.distance.index(min_dist)
        min_id = np.argmin(self.distance)
        min_dist = self.distance[min_id]
        return min_dist, min_id

    def __len__(self):
        assert(len(self.distance) == len(self.predecessor))
        return len(self.distance)
    #
    # def __repr__(self):
    #     return 'min dist: {0}, min pred id: {1}'.format(self.extract_min())

class DAGSearch(object):
    def __init__(self, graph):
        assert(isinstance(graph, LadderGraph))
        if graph.size == 0:
            raise ValueError('input ladder graph is empty!')
        self.graph = graph
        self.solution = [SolutionRung() for i in range(graph.get_rungs_size())]

        # allocate everything we need
        for i in range(graph.get_rungs_size()):
            n_verts = graph.get_rung_vert_size(i)
            assert(n_verts > 0)
            self.solution[i].distance = np.zeros(n_verts)
            self.solution[i].predecessor = np.zeros(n_verts, dtype=int)

    @classmethod
    def from_ladder_graph(cls, graph):
        return cls(graph)

    def distance(self, r_id, v_id):
        return self.solution[r_id].distance[v_id]

    def predecessor(self, r_id, v_id):
        return self.solution[r_id].predecessor[v_id]

    def run(self):
        """forward cost propagation"""
        if len(self.solution) == 0:
            raise ValueError('The initial solution is empty!')

        # TODO: add st_conf cost to SolutionRung 0
        # * first rung init to 0
        self.solution[0].distance = np.zeros(len(self.solution[0]))
        # * other rungs init to inf
        for j in range(1, len(self.solution)):
            self.solution[j].distance = np.inf*np.ones(len(self.solution[j]))

        for r_id in range(0, len(self.solution)-1):
            n_verts = self.graph.get_rung_vert_size(r_id)
            next_r_id = r_id + 1
            # for each vert in the out edge list
            # min_rung_cost = np.inf
            for v_id in range(n_verts):
                u_cost = self.distance(r_id, v_id)
                # outgoing edges
                edges = self.graph.get_edges(r_id)[v_id]
                # find the minimal distance edge in all the outgoing edges from v_id
                dv = np.inf
                for edge in edges:
                    dv = u_cost + edge.cost
                    # print('out v# ', edge.idx, ', dist ', self.distance(next_r_id, edge.idx), ', dv: ', dv)
                    if dv < self.distance(next_r_id, edge.idx):
                        self.solution[next_r_id].distance[edge.idx] = dv
                        self.solution[next_r_id].predecessor[edge.idx] = v_id
                    # update the minimal distance for the current rung
                    # if dv < min_rung_cost:
                    #     min_rung_cost = dv
            # print('Rung #{}: {}'.format(r_id, min_rung_cost))

        return min(self.solution[-1].distance)

    def shortest_path(self):
        if len(self.solution) == 0:
            # TODO: more detailed checks
            raise ValueError('The initial solution is empty!')

        _, min_val_id = self.solution[-1].extract_min()
        path_idx = np.zeros(len(self.solution), dtype=int)

        current_v_id = min_val_id
        count = len(path_idx) - 1
        while count >= 0:
            path_idx[count] = current_v_id
            current_v_id = self.predecessor(count, current_v_id)
            count -= 1

        sol = []
        for r_id, v_id in enumerate(path_idx):
            data = self.graph.get_vert_data(r_id, v_id)
            sol.append(data)

        return sol



from random import random
from time import time

from .utils import INF, argmin, elapsed_time

__all__ = [
    'rrt_star',
    'informed_rrt_star',
    ]

EPSILON = 1e-6
PRINT_FREQUENCY = 100

class OptimalNode(object):

    def __init__(self, config, parent=None, d=0, path=[], iteration=None):
        self.config = config
        self.parent = parent
        self.children = set()
        self.d = d
        self.path = path
        if parent is not None:
            self.cost = parent.cost + d
            self.parent.children.add(self)
        else:
            self.cost = d
        self.solution = False
        self.creation = iteration
        self.last_rewire = iteration

    def set_solution(self, solution):
        if self.solution is solution:
            return
        self.solution = solution
        if self.parent is not None:
            self.parent.set_solution(solution)

    def retrace(self):
        if self.parent is None:
            return self.path + [self.config]
        return self.parent.retrace() + self.path + [self.config]

    def rewire(self, parent, d, path, iteration=None):
        if self.solution:
            self.parent.set_solution(False)
        self.parent.children.remove(self)
        self.parent = parent
        self.parent.children.add(self)
        if self.solution:
            self.parent.set_solution(True)
        self.d = d
        self.path = path
        self.update()
        self.last_rewire = iteration

    def update(self):
        self.cost = self.parent.cost + self.d
        for n in self.children:
            n.update()

    def clear(self):
        self.node_handle = None
        self.edge_handle = None

    def draw(self, draw_fn, valid=True, remove_all=False):
        segment = [] if self.parent is None else [self.config, self.parent.config]
        draw_fn(self.config, segment, valid, valid, remove_all)
        # from manipulation.primitives.display import draw_node, draw_edge
        # color = (0, 0, 1, .5) if self.solution else (1, 0, 0, .5)
        # self.node_handle = draw_node(env, self.config, color=color)
        # if self.parent is not None:
        #     self.edge_handle = draw_edge(
        #         env, self.config, self.parent.config, color=color)

    def __str__(self):
        return self.__class__.__name__ + '(' + str(self.config) + ')'
    __repr__ = __str__


def safe_path(sequence, collision):
    path = []
    for q in sequence:
        if collision(q):
            break
        path.append(q)
    return path

##################################################

def rrt_star(start, goal, distance_fn, sample_fn, extend_fn, collision_fn, radius,
             max_time=INF, max_iterations=INF, goal_probability=.2, informed=True, verbose=False, draw_fn=None):
    """
    :param start: Start configuration - conf
    :param goal: End configuration - conf
    :param distance_fn: Distance function - distance_fn(q1, q2)->float
    :param sample_fn: Sample function - sample_fn()->conf
    :param extend_fn: Extension function - extend_fn(q1, q2)->[q', ..., q"]
    :param collision_fn: Collision function - collision_fn(q)->bool
    :param max_time: Maximum runtime - float
    :return: Path [q', ..., q"] or None if unable to find a solution
    """
    if collision_fn(start) or collision_fn(goal):
        return None
    nodes = [OptimalNode(start)]
    goal_n = None
    start_time = time()
    iteration = 0
    while (elapsed_time(start_time) < max_time) and (iteration < max_iterations):
        do_goal = goal_n is None and (iteration == 0 or random() < goal_probability)
        s = goal if do_goal else sample_fn()
        # Informed RRT*
        if informed and (goal_n is not None) and (distance_fn(start, s) + distance_fn(s, goal) >= goal_n.cost):
            continue
        if iteration % PRINT_FREQUENCY == 0:
            success = goal_n is not None
            cost = goal_n.cost if success else INF
            if verbose:
                print('Iteration: {} | Time: {:.3f} | Success: {} | {} | Cost: {:.3f}'.format(
                    iteration, elapsed_time(start_time), success, do_goal, cost))
        iteration += 1

        nearest = argmin(lambda n: distance_fn(n.config, s), nodes)
        path = safe_path(extend_fn(nearest.config, s), collision_fn)
        if len(path) == 0:
            continue
        new = OptimalNode(path[-1], parent=nearest, d=distance_fn(
            nearest.config, path[-1]), path=path[:-1], iteration=iteration)
        # if safe and do_goal:
        if do_goal and (distance_fn(new.config, goal) < EPSILON):
            goal_n = new
            goal_n.set_solution(True)
        # TODO - k-nearest neighbor version
        neighbors = filter(lambda n: distance_fn(n.config, new.config) < radius, nodes)
        nodes.append(new)

        # TODO: smooth solution once found to improve the cost bound
        for n in neighbors:
            d = distance_fn(n.config, new.config)
            if (n.cost + d) < new.cost:
                path = safe_path(extend_fn(n.config, new.config), collision_fn)
                if (len(path) != 0) and (distance_fn(new.config, path[-1]) < EPSILON):
                    new.rewire(n, d, path[:-1], iteration=iteration)
        for n in neighbors:  # TODO - avoid repeating work
            d = distance_fn(new.config, n.config)
            if (new.cost + d) < n.cost:
                path = safe_path(extend_fn(new.config, n.config), collision_fn)
                if (len(path) != 0) and (distance_fn(n.config, path[-1]) < EPSILON):
                    n.rewire(new, d, path[:-1], iteration=iteration)

        if draw_fn:
            nodes[0].draw(draw_fn, remove_all=True)
            for n in nodes[1:]:
                n.draw(draw_fn, remove_all=False)

    if goal_n is None:
        return None
    return goal_n.retrace()

def informed_rrt_star(start, goal, distance_fn, sample_fn, extend_fn, collision_fn, radius, **kwargs):
    return rrt_star(start, goal, distance_fn, sample_fn, extend_fn, collision_fn, radius, informed=True, **kwargs)



"""
********************************************************************************
motion_planners
********************************************************************************

.. currentmodule:: pybullet_planning.motion_planners

Python implementations of several robotic motion planners. This is a fork of Caelan's
`motion-planners <https://github.com/caelan/motion-planners>`_ repo, which is designed
to be flexible and independent of simulation platforms. `pybullet_planning` includes
this package as a built-in component as it is frequently used.

We are now up-to-date with commit `e6f23053e <https://github.com/caelan/motion-planners/commit/e6f23053e441af091b898b7f56c6fee48223be48>`_ of ``motion-planners``.

Sampling-based:

- Probabilistic Roadmap (PRM)
- Rapidly-Exploring Random Tree (RRT)
- RRT-Connect (BiRRT)
- Linear Shortcutting
- MultiRRT
- RRT*

Grid search:

- Breadth-First Search (BFS)
- A*
- Ladder graph DAG Search

Probabilistic Roadmap (PRM)
----------------------------

.. currentmodule:: pybullet_planning.motion_planners.lazy_prm

.. autosummary::
    :toctree: generated/
    :nosignatures:

    lazy_prm

RRT-Connect (BiRRT)
----------------------------

.. currentmodule:: pybullet_planning.motion_planners.rrt_connect

.. autosummary::
    :toctree: generated/
    :nosignatures:

    rrt_connect
    birrt

Smoothing
---------------------------

.. currentmodule:: pybullet_planning.motion_planners.smoothing

.. autosummary::
    :toctree: generated/
    :nosignatures:

    rrt_connect

"""

from .prm import *
from .lazy_prm import *
from .rrt_connect import *
from .rrt import *
from .rrt_star import *
from .lattice import *
from .smoothing import *
from .meta import *
from .utils import *

__all__ = [name for name in dir() if not name.startswith('_')]
